/** @file emu.cpp

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.08.18 -

	@note
	Modified for BML3MK5 by Sasaji at 2011.06.17

	@brief [ win32 emulation i/f ]
*/

#include "emu.h"

#include "config.h"

// ----------------------------------------------------------------------------
// initialize
// ----------------------------------------------------------------------------

EMU::EMU()
{
	// get module path
	GetModuleFileName(NULL, app_path, _MAX_PATH);
	int pt = (int)_tcslen(app_path);
	while(pt >= 0 && app_path[pt] != _T('\\')) {
		pt--;
	}
	app_path[pt + 1] = _T('\0');

	vm = NULL;
	ledbox = NULL;
	configbox = NULL;
	keybindbox = NULL;
	msgboard = NULL;

#ifdef USE_DIRECT3D
	pD3Device = NULL;
	pD3Dmixsuf = NULL;
	pD3Dsource = NULL;
	pD3Dorigin = NULL;
#ifdef USE_SCREEN_ROTATE
	pD3Drotate = NULL;
#endif
	lpD3DBmp = NULL;
#endif

	EMU_SCREEN();

	// open logfile
	open_logfile();
	out_log(LOG_INFO, _T("START [%s]"), app_path);

	//
	vm_pause = vm_syspause = vm_usrpause = false;
}

void EMU::initialize(HWND hwnd, HINSTANCE hinst)
{
	// store main window handle
	main_window_handle = hwnd;
	instance_handle = hinst;

#ifdef USE_CPU_CLOCK_LOW
	cpu_clock_low = config.cpu_clock_low;
#endif

#ifdef _BML3MK5
	// create dialogbox
	ledbox = new LedBox(hinst);
	configbox = new ConfigBox(hinst, this, &config);
	keybindbox = new KeybindBox(hinst);
	volumebox = new VolumeBox(hinst, this, &config);
	msgboard = new MsgBoard(hwnd, this);
	msgboard->SetVisible(config.show_dlg & 2 ? true : false);
#endif

	// initialize
	vm = new VM(this);
	initialize_input();
	initialize_screen();
	initialize_sound();
#ifdef USE_FD1
	initialize_disk_insert();
#endif
#ifdef USE_MEDIA
	initialize_media();
#endif
#ifdef USE_SOCKET
	initialize_socket();
#endif
//	vm->reset();
}

EMU::~EMU()
{
	out_log(LOG_INFO, _T("EXIT"));
	close_logfile();

	if (msgboard) {
		delete msgboard;
	}
}

void EMU::release()
{
	PEXCEPTION_POINTERS Ex = NULL;

	if (ledbox) {
		delete ledbox;
	}
	if (volumebox) {
		delete volumebox;
	}
	if (keybindbox) {
		delete keybindbox;
	}
	if (configbox) {
		delete configbox;
	}
	__try {
		if(vm) {
			delete vm;
		}
	}
	__except (Ex = GetExceptionInformation(),
			EXCEPTION_EXECUTE_HANDLER) {
		out_log(LOG_ERROR,_T("EMU::release Exception: %x"),Ex->ExceptionRecord->ExceptionCode);
	}

	release_input();
	release_screen();
	release_sound();
#ifdef USE_MEDIA
	release_media();
#endif
#ifdef USE_SOCKET
	release_socket();
#endif
}

// ----------------------------------------------------------------------------
// drive machine
// ----------------------------------------------------------------------------

int EMU::frame_interval()
{
#ifdef SUPPORT_VARIABLE_TIMING
	return (int)(1024. * 1000. / vm->frame_rate() + 0.5);

#else
	return (int)(1024. * 1000. / FRAMES_PER_SEC + 0.5);
#endif
}

int EMU::run(int split_num)
{
	update_input();
	update_timer();
#ifdef USE_FD1
	update_disk_insert();
#endif
#ifdef USE_SOCKET
	update_socket();
#endif

	// virtual machine may be driven to fill sound buffer
	int extra_frames = 0;
	update_sound(&extra_frames);

	// drive virtual machine
	if(extra_frames == 0) {
		vm->run(split_num);
		extra_frames = 1;
	}
#ifdef _BML3MK5
	if (msgboard && split_num == 0) {
		msgboard->CountDown();
	}
#endif
	return extra_frames;
}

// update status if need
void EMU::update()
{
	update_timer();
#ifdef USE_FD1
	update_disk_insert();
#endif
#ifdef USE_SOCKET
	update_socket();
#endif

	// virtual machine may be driven to fill sound buffer
	int extra_frames = 0;
	update_sound(&extra_frames);

	return;
}

void EMU::reset()
{
	// update config
	if (configbox) {
		configbox->Update();
	}
#ifdef USE_CPU_CLOCK_LOW
	if(cpu_clock_low != config.cpu_clock_low) {
		// stop sound
		if(sound_ok && sound_started) {
			lpdsb->Stop();
			sound_started = false;
		}
		// reinitialize virtual machine
		delete vm;
		vm = new VM(this);
		vm->initialize_sound(sound_rate, sound_samples);
		vm->reset();

		// restore inserted floppy disks
#ifdef USE_FD1
		for(int drv = 0; drv < 8; drv++) {
			if(disk_insert[drv].path[0] != _T('\0')) {
				vm->open_disk(drv, disk_insert[drv].path, disk_insert[drv].offset);
			}
		}
#endif
		cpu_clock_low = config.cpu_clock_low;
	}
	else {
#endif
		// reset virtual machine
		vm->reset();
#ifdef USE_CPU_CLOCK_LOW
	}
#endif

#ifdef USE_MEDIA
	stop_media();
#endif
	// restart recording
//	restart_rec_video();
//	restart_rec_sound();
}

#ifdef USE_SPECIAL_RESET
void EMU::special_reset()
{
	// reset virtual machine
	vm->special_reset();
#ifdef USE_MEDIA
	stop_media();
#endif
	// restart recording
//	restart_rec_video();
//	restart_rec_sound();
}

void EMU::warm_reset()
{
	// reset virtual machine
	vm->warm_reset();
#ifdef USE_MEDIA
	stop_media();
#endif
}
#endif

void EMU::change_dipswitch(int num)
{
	config.dipswitch ^= (1 << num);
	vm->change_dipswitch(num);
}

#ifdef USE_POWER_OFF
void EMU::notify_power_off()
{
	vm->notify_power_off();
}
#endif

bool EMU::get_pause(int idx) {
	switch(idx) {
		case 1:
			return vm_syspause;
			break;
		case 2:
			return vm_usrpause;
			break;
	}
	return vm_pause;
}
void EMU::set_pause(int idx, bool val) {
	switch(idx) {
		case 1:
			vm_syspause = val;
			break;
		case 2:
			vm_usrpause = val;
			out_info(_T("Pause"), val, -1);

			break;
	}
	vm_pause = (vm_syspause | vm_usrpause);

	// mute sound
	mute_sound(vm_pause);
}
bool *EMU::get_pause_ptr(void) {
	return &vm_pause;
}

void EMU::application_path(_TCHAR* path)
{
	_tcscpy(path, app_path);
}

_TCHAR *EMU::application_path()
{
	return app_path;
}

// ----------------------------------------------------------------------------
// logfile
// ----------------------------------------------------------------------------

void EMU::open_logfile()
{
#ifdef _LOG_CONSOLE
	AllocConsole();
	hConsole = GetStdHandle(STD_OUTPUT_HANDLE);
	SetConsoleTitle("Debug Log");
#endif
#ifdef _LOG_FILE
	// get application path
	_TCHAR log_path[_MAX_PATH];
	_stprintf(log_path, _T("%s%s.log"), app_path, _T(CONFIG_NAME));

	fplog = _tfopen(log_path, _T("w"));
#endif
}

void EMU::close_logfile()
{
#ifdef _LOG_CONSOLE
	FreeConsole();
#endif
#ifdef _LOG_FILE
	if(fplog) {
		fclose(fplog);
	}
#endif
}

void EMU::out_logv(int level, const _TCHAR* format, va_list ap)
{
	_TCHAR buffer[1024];
	_TCHAR levelstr[32];

	_vstprintf(buffer, format, ap);

	switch(level) {
	case LOG_WARN:
		_tcscpy(levelstr, _T("WARN"));
		break;
	case LOG_ERROR:
		_tcscpy(levelstr, _T("ERROR"));
		break;
	case LOG_DEBUG:
		_tcscpy(levelstr, _T("DEBUG"));
		break;
	default:
		_tcscpy(levelstr, _T("INFO"));
		break;
	}

#ifdef _LOG_CONSOLE
	DWORD dwWritten;
	WriteConsole(hConsole, buffer, _tcslen(buffer), &dwWritten, NULL);
#endif
#ifdef _LOG_FILE
	if(fplog) {
		SYSTEMTIME st;
		GetLocalTime(&st);

		_ftprintf(fplog, _T("%04d-%02d-%02d %02d:%02d:%02d.%03d %-5s: %s\n")
			, st.wYear, st.wMonth, st.wDay, st.wHour, st.wMinute, st.wSecond, st.wMilliseconds
			, levelstr, buffer);
		fflush(fplog);
	}
#endif
#ifdef _BML3MK5
	if (msgboard && level != LOG_DEBUG) {
		msgboard->SetMessageF(_T("[%s] %s"),levelstr,buffer);
	}
#endif
}

void EMU::out_log(int level, const _TCHAR* format, ...)
{
	va_list ap;

	va_start(ap, format);
	out_logv(level, format, ap);
	va_end(ap);
}

void EMU::out_debug(const _TCHAR* format, ...)
{
#ifdef _DEBUG_LOG
	va_list ap;

	va_start(ap, format);
	out_logv(LOG_DEBUG, format, ap);
	va_end(ap);
#endif
}

// ----------------------------------------------------------------------------

void EMU::out_msg(const _TCHAR* msg, bool set, int sec)
{
#ifdef _BML3MK5
	if (msgboard) {
		if (set) {
			msgboard->SetMessage(msg, sec);
		} else {
			msgboard->DeleteMessage(msg);
		}
	}
#endif
}
void EMU::out_msg(const _TCHAR* msg, bool set)
{
#ifdef _BML3MK5
	if (msgboard) {
		if (set) {
			msgboard->SetMessage(msg);
		} else {
			msgboard->DeleteMessage(msg);
		}
	}
#endif
}

void EMU::out_info(const _TCHAR* msg, bool set, int sec)
{
#ifdef _BML3MK5
	if (msgboard) {
		if (set) {
			msgboard->SetInfo(msg, sec);
		} else {
			msgboard->DeleteInfo(msg);
		}
	}
#endif
}

void EMU::out_info(const _TCHAR* msg, bool set)
{
#ifdef _BML3MK5
	if (msgboard) {
		if (set) {
			msgboard->SetInfo(msg);
		} else {
			msgboard->DeleteInfo(msg);
		}
	}
#endif
}

void EMU::out_infof(const _TCHAR* format, ...)
{
#ifdef _BML3MK5
	if (msgboard) {
		_TCHAR buffer[1024];

		va_list ap;

		va_start(ap, format);
		_vstprintf(buffer, format, ap);
		va_end(ap);

		msgboard->SetInfo(buffer);
	}
#endif
}

// ----------------------------------------------------------------------------
// user interface
// ----------------------------------------------------------------------------

void EMU::change_cpu_power(int num)
{
	config.cpu_power = num;

	if (num == 0) {
		out_infof(_T("CPU x0.5"));
	} else {
		out_infof(_T("CPU x%d"), (1 << (num - 1)));
	}
	update_config();
}
#ifdef _BML3MK5
void EMU::change_sync_irq()
{
	config.sync_irq = !config.sync_irq;

	out_infof(_T("Sync IRQ %s"),(config.sync_irq ? "ON" : "OFF"));
	update_config();
}
#endif
#ifdef USE_CART
void EMU::open_cart(_TCHAR* file_path)
{
	vm->open_cart(file_path);

	// restart recording
	restart_rec_video();
	restart_rec_sound();
}
void EMU::close_cart()
{
	vm->close_cart();

	// stop recording
	stop_rec_video();
	stop_rec_sound();
}
#endif

#ifdef USE_FD1
bool EMU::open_disk(int drv, _TCHAR* file_path, int offset, uint32 flags)
{
	if(vm->disk_inserted(drv) || disk_insert[drv].wait_count != 0) {
		if(vm->disk_inserted(drv)) {
			vm->close_disk(drv);
#ifdef _BML3MK5
		}
		return vm->open_disk(drv, file_path, offset, flags);
#else
			// wait 0.5sec
#ifdef SUPPORT_VARIABLE_TIMING
			disk_insert[drv].wait_count = (int)(vm->frame_rate() / 2);
#else
			disk_insert[drv].wait_count = (int)(FRAMES_PER_SEC / 2);
#endif
		}
		_tcscpy(disk_insert[drv].path, file_path);
		disk_insert[drv].offset = offset;
		disk_insert[drv].flags = flags;

		return true;
#endif
	}
	else {
		return vm->open_disk(drv, file_path, offset, flags);
	}
}
bool EMU::open_disk(int drv, _TCHAR* file_path, int offset)
{
	return open_disk(drv, file_path, offset, 0);
}
void EMU::close_disk(int drv)
{
	vm->close_disk(drv);
}
void EMU::initialize_disk_insert()
{
	memset(disk_insert, 0, sizeof(disk_insert));
}
void EMU::update_disk_insert()
{
	for(int drv = 0; drv < 8; drv++) {
		if(disk_insert[drv].wait_count != 0 && --disk_insert[drv].wait_count == 0) {
			vm->open_disk(drv, disk_insert[drv].path, disk_insert[drv].offset, disk_insert[drv].flags);
		}
	}
}
int EMU::change_disk(int drv)
{
	return vm->change_disk(drv);
}
int EMU::get_disk_side(int drv)
{
	return vm->get_disk_side(drv);
}
#endif

#ifdef USE_QUICKDISK
void EMU::open_quickdisk(_TCHAR* file_path)
{
	vm->open_quickdisk(file_path);
}
void EMU::close_quickdisk()
{
	vm->close_quickdisk();
}
#endif

#ifdef USE_DATAREC
bool EMU::play_datarec(_TCHAR* file_path)
{
	return vm->play_datarec(file_path);
}
bool EMU::rec_datarec(_TCHAR* file_path)
{
	return vm->rec_datarec(file_path);
}
void EMU::close_datarec()
{
	vm->close_datarec();
}
void EMU::rewind_datarec()
{
	vm->rewind_datarec();
}
void EMU::fast_forward_datarec()
{
	vm->fast_forward_datarec();
}
void EMU::realmode_datarec()
{
	vm->realmode_datarec();
}
#endif

#ifdef USE_DATAREC_BUTTON
void EMU::push_play()
{
	vm->push_play();
}
void EMU::push_stop()
{
	vm->push_stop();
}
#endif

#ifdef USE_BINARY_FILE1
void EMU::load_binary(int drv, _TCHAR* file_path)
{
	vm->load_binary(drv, file_path);
}
void EMU::save_binary(int drv, _TCHAR* file_path)
{
	vm->save_binary(drv, file_path);
}
#endif
#ifdef USE_SCANLINE
void EMU::change_screen_scanline()
{
	config.scan_line = !config.scan_line;

	out_infof(_T("Scanline %s"),(config.scan_line ? "ON" : "OFF"));
	update_config();
}
#endif
#ifdef _BML3MK5
void EMU::change_screen_use_direct3d(int num)
{
	if(enable_direct3d) {
		uint8 pre_num = config.use_direct3d;

		if (num >= 0) {
			config.use_direct3d = (pre_num == num) ? 0 : num;
		} else {
			config.use_direct3d = (pre_num + 1) % 3;
		}

		if (config.use_direct3d != 0) {
			set_d3dpresent_interval();
			reset_d3device(main_window_handle);
		}

		if (pD3Device != NULL && pre_num == 0 && config.use_direct3d != 0) {
			copy_d3dsuf_dib(pD3Dorigin, lpBmp, false);
		} else if (pre_num != 0 && config.use_direct3d == 0) {
			copy_d3dsuf_dib(pD3Dorigin, lpBmp, true);
		}

		switch (config.use_direct3d) {
			case 0:
				out_infof(_T("Direct3D OFF"));
				break;
			case 1:
				out_infof(_T("Direct3D ON (Sync)"));
				break;
			case 2:
				out_infof(_T("Direct3D ON (Async)"));
				break;
		}
		update_config();
	}
}
void EMU::change_screen_afterimage(int num)
{
	if (num >= 0) {
		config.afterimage = (config.afterimage == num) ? 0 : num;
	} else {
		config.afterimage = (config.afterimage + 1) % 3;
	}

	if (config.afterimage != 0) {
		out_infof(_T("Afterimage%d ON"), config.afterimage);
	} else {
		out_info(_T("Afterimage OFF"));
	}
	update_config();
}
void EMU::change_screen_keepimage(int num)
{
	if (num >= 0) {
		config.keepimage = (config.keepimage == num) ? 0 : num;
	} else {
		config.keepimage = (config.keepimage + 1) % 3;
	}

	if (config.keepimage != 0) {
		out_infof(_T("Keepimage%d ON"), config.keepimage);
	} else {
		out_info(_T("Keepimage OFF"));
	}
	update_config();
}
#endif

bool EMU::now_skip()
{
	return (bool)vm->now_skip();
}

void EMU::update_config()
{
	vm->update_config();
}

#ifdef USE_PRINTER
void EMU::save_printer(int drv, _TCHAR* file_path)
{
	vm->save_printer(drv, file_path);
}
void EMU::clear_printer(int drv)
{
	vm->clear_printer(drv);
}
int  EMU::get_printer_buffer_size(int drv)
{
	return vm->get_printer_buffer_size(drv);
}
uint8* EMU::get_printer_buffer(int drv)
{
	return vm->get_printer_buffer(drv);
}
#if 0
void EMU::send_printer_data(uint8 *datas, int len)
{
	HWND hTargetWnd = NULL;
	COPYDATASTRUCT cp_data;
	LRESULT lr;

	hTargetWnd = FindWindow(_T("MPPRINTER"), NULL);

	if (hTargetWnd != NULL && len > 0) {
		cp_data.dwData = 1;
		cp_data.cbData = len;
		cp_data.lpData = (PVOID)datas;

		lr = SendMessage(hTargetWnd, WM_COPYDATA, (WPARAM)main_window_handle, (LPARAM)&cp_data);
	}
}
#endif
void EMU::enable_printer_direct(int drv)
{
	vm->enable_printer_direct(drv);
}
bool EMU::print_printer(int drv)
{
	return vm->print_printer(drv);
}
#endif

#ifdef _BML3MK5
#if 0
static HWND another_main_window_handle = NULL;

BOOL CALLBACK SearchOurWindowsProc(HWND hWnd, LPARAM lParam)
{
	_TCHAR sClass[100];
	HWND hMainWnd = NULL;

	hMainWnd = (HWND)lParam;

	// search another window
	if (hWnd != NULL) {
		GetClassName(hWnd, sClass, 100);

		if (_tcscmp(sClass,_T("BML3MK5")) == 0 && hWnd != hMainWnd) {
			another_main_window_handle = hWnd;
			return FALSE;
		}
	}
	return TRUE;
}
#endif

void EMU::enable_comm_server(int drv)
{
	vm->enable_comm_server(drv);
}
void EMU::enable_comm_connect(int drv)
{
	vm->enable_comm_connect(drv);
}

#if 0
void EMU::send_comm_data(uint8 *data, size_t size)
{
	COPYDATASTRUCT cp_data;
	LRESULT lr;

	if (another_main_window_handle == NULL) {
		EnumWindows(SearchOurWindowsProc, (LPARAM)main_window_handle);
	}
	if (another_main_window_handle != NULL) {
		cp_data.dwData = 1;
		cp_data.cbData = (DWORD)size;
		cp_data.lpData = (void *)data;
		lr = SendMessage(another_main_window_handle, WM_COPYDATA, (WPARAM)main_window_handle, (LPARAM)&cp_data);
		if (lr == FALSE) {
			another_main_window_handle = NULL;
		}
	}
}

void EMU::recv_comm_data(uint8 *data, size_t size)
{
	vm->recv_comm_data(data, size);
}
#endif

int EMU::update_led()
{
	if (vm) {
		return vm->update_led();
	}
	return 0;
}

void EMU::show_led()
{
	if (ledbox) {
		config.show_dlg ^= 1;
		ledbox->Show(config.show_dlg & 9);
		out_infof(_T("%s LED"), (config.show_dlg & 1) ? "Show" : "Hide");
	}
}

void EMU::inside_led()
{
	if (ledbox) {
		config.show_dlg ^= 8;
		ledbox->Show(config.show_dlg & 9);
		out_infof(_T("%s LED"), (config.show_dlg & 8) ? "Inside" : "Outside");
	}
}

void EMU::move_led()
{
	if (ledbox) {
		ledbox->Move();
	}
}

bool EMU::is_shown_ledbox(int id)
{
	switch(id) {
		case 0:
			return (config.show_dlg & 1) ? true : false;
			break;
		case 1:
			return (config.show_dlg & 8) ? true : false;
			break;
	}
	return false;
}

void EMU::show_message_board()
{
	if (msgboard) {
		config.show_dlg ^= 2;
		msgboard->SetVisible(config.show_dlg & 2 ? true : false);
		out_infof(_T("%s Message"), (config.show_dlg & 2) ? "Show" : "Hide");
	}
}

bool EMU::is_shown_message_board()
{
	return (config.show_dlg & 2) ? true : false;
}

void EMU::toggle_use_joypad()
{
	config.show_dlg ^= 4;
	if (config.show_dlg & 4) {
		use_joystick = true;
		reset_joystick();
		out_infof(_T("Enable Joypad"));
	} else {
		release_joystick();
		use_joystick = false;
		out_infof(_T("Disable Joypad"));
	}
}

bool EMU::is_enable_joypad()
{
	return (config.show_dlg & 4) ? true : false;
}

void EMU::show_config_dialog()
{
	set_pause(1, true);
	if (configbox->Show(main_window_handle) == IDOK) {
		save_config();
	}
	set_pause(1, false);
}

void EMU::show_keybind_dialog()
{
	set_pause(1, true);
	if (keybindbox->Show(main_window_handle) == IDOK) {
		vm->save_keybind();
		if (use_joystick) {
			reset_joystick();
		}
	}
	set_pause(1, false);
}

void EMU::save_state(_TCHAR* filename)
{
	vm->save_state(filename);
}

void EMU::load_state(_TCHAR* filename)
{
	vm->load_state(filename);
}

int EMU::get_parami(enumParamId id)
{
	switch(id) {
		case ConfigBoxFddType:
			return configbox->GetParam(FddType);
			break;
		case ConfigBoxWavReverse:
			return configbox->GetParam(WavReverse);
			break;
		case ConfigBoxWavHalf:
			return configbox->GetParam(WavHalf);
			break;
		case ConfigBoxWavCorrect:
			return configbox->GetParam(WavCorrect);
			break;
		case ConfigBoxWavSampleRate:
			return configbox->GetParam(WavSampleRate);
			break;
		case ConfigBoxWavSampleBits:
			return configbox->GetParam(WavSampleBits);
			break;
		case ConfigBoxIOPort:
			return configbox->GetParam(IOPort);
			break;
	}
	return 0;
}

#endif	/* _BML3MK5 */
