/** @file emu.h

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.08.18 -

	@note
	Modified for BML3MK5 by Sasaji at 2011.06.17

	@brief [ win32 emulation i/f ]
*/

#ifndef _EMU_H_
#define _EMU_H_

// DirectX
#define DIRECTSOUND_VERSION 0x900
#ifdef USE_DIRECT3D
//#define DIRECT3D_VERSION 0x900
#include <d3d9.h>
#pragma comment(lib, "d3d9.lib")
#ifdef USE_DIRECT3DX
#include <d3dx9.h>
#pragma comment(lib, "d3dx9.lib")
#endif
#endif

// for log
// #define _LOG_CONSOLE
#define _LOG_FILE

#define LOG_INFO	0
#define LOG_WARN	1
#define LOG_ERROR	2
#define LOG_DEBUG	3

// for debug
//#define _DEBUG_LOG
#ifdef _DEBUG_LOG
	// output debug log to console
//	#define _DEBUG_CONSOLE
	// output debug log to file
	#define _DEBUG_FILE

	// output cpu debug log
//	#define _CPU_DEBUG_LOG
	// output fdc debug log
//	#define _FDC_DEBUG_LOG
	// output i/o debug log
//	#define _IO_DEBUG_LOG
#endif

#include <windows.h>
#include <windowsx.h>
#include <mmsystem.h>
#include <commdlg.h>
#include <commctrl.h>
#include <stdio.h>
#include "common.h"
#include "vm/vm.h"
#include "res/resource.h"
#include "ledbox.h"
#include "configbox.h"
#include "keybindbox.h"
#include "volumebox.h"
#include "msgboard.h"

#define WM_RESIZE  (WM_USER + 1)
#define WM_SOCKET0 (WM_USER + 2)
#define WM_SOCKET1 (WM_USER + 3)
#define WM_SOCKET2 (WM_USER + 4)
#define WM_SOCKET3 (WM_USER + 5)
#define WM_SOCKET4 (WM_USER + 6)
#define WM_SOCKET5 (WM_USER + 7)

#ifndef SCREEN_WIDTH_ASPECT
#define SCREEN_WIDTH_ASPECT SCREEN_WIDTH
#endif
#ifndef SCREEN_HEIGHT_ASPECT
#define SCREEN_HEIGHT_ASPECT SCREEN_HEIGHT
#endif
#ifndef WINDOW_WIDTH
#define WINDOW_WIDTH SCREEN_WIDTH_ASPECT
#endif
#ifndef WINDOW_HEIGHT
#define WINDOW_HEIGHT SCREEN_HEIGHT_ASPECT
#endif

#include <dsound.h>
#include <vfw.h>

#ifdef USE_SOCKET
#include <winsock.h>
#endif

// check memory leaks
#ifdef _DEBUG
#define _CRTDBG_MAP_ALLOC
#include <crtdbg.h>
#define malloc(s) _malloc_dbg(s, _NORMAL_BLOCK, __FILE__, __LINE__)
#define new new(_NORMAL_BLOCK, __FILE__, __LINE__)
#endif

#ifdef USE_MEDIA
#define MEDIA_MAX 64
#endif
#ifdef USE_SOCKET
#define SOCKET_MAX 6
#define SOCKET_BUFFER_MAX 0x100000
#endif

#define KEY_MOD_ALT_KEY 0x0001

class FIFO;
class FILEIO;

/**
	@brief emulation i/f
*/
class EMU
{
protected:
	VM* vm;
private:
	// ----------------------------------------
	// input
	// ----------------------------------------
	/// @name input private methods
	//@{
	void initialize_input();
	void release_input();
	void update_input();

	void initialize_joystick();
	void release_joystick();
	//@}
	/// @name input private members
	//@{
#ifdef USE_KEYCODE_CONV
	uint8 keycode_conv[256];
#endif
	uint8 key_status[256];	// windows key code mapping
#ifdef USE_SHIFT_NUMPAD_KEY
	uint8 key_converted[256];
	bool key_shift_pressed, key_shift_released;
#endif
	bool lost_focus;
	int  key_mod;

	uint32 joy_status[2];	// joystick #1, #2 (b0 = up, b1 = down, b2 = left, b3 = right, b4-b27 = trigger #1-#24
	int joy_num;
	int joyid_map[2];
	uint32 joy_xmin[2], joy_xmax[2];
	uint32 joy_ymin[2], joy_ymax[2];
	bool joy_enabled[2];
	bool use_joystick;
	INPUT dummy_key;
	int send_dummy_key;

	int mouse_status[3];	// x, y, button (b0 = left, b1 = right)
	bool mouse_enabled;

#ifdef USE_AUTO_KEY
	FIFO* autokey_buffer;
	int autokey_phase, autokey_shift;
	uint8 autokey_code;
#ifdef _BML3MK5
	int autokey_graph;
#endif
#endif
	//@}
public:
	/// @name input public methods
	//@{
	void reset_joystick();
	void update_joystick();
	//@}

private:
	// ----------------------------------------
	// screen
	// ----------------------------------------
	/// @name screen private methods
	//@{
	void EMU_SCREEN();
	void initialize_screen();
	void release_screen();
	void create_dib_section(HDC hdc, int width, int height, HDC *hdcDib, HBITMAP *hBmp, HBITMAP *hOldBmp, LPBYTE *lpBuf, scrntype **lpBmp, LPBITMAPINFO *lpDib);
	void create_dib_rec_video(HDC hdc, int new_size);
	void copy_dib_rec_video();
	BOOL get_monitor_device(HMONITOR hMonitor, HDC hdcMonitor, LPRECT lprcMonitor);

#ifdef USE_DIRECT3D
	void initialize_d3device(HWND hWnd);
	HRESULT create_d3device(HWND hWnd);
	void release_d3device();
	void terminate_d3device();

	void update_screen_dc(HDC hdc);
	void update_screen_d3d();
	inline void copy_d3dsuf_dib(PDIRECT3DSURFACE9 suf, scrntype *buf, bool to_dib);
	void create_d3dofflinesurface();
	HRESULT reset_d3device(HWND hWnd);
	void set_d3dpresent_interval();
#endif
	//@}
	/// @name screen private static methods
	//@{
	static int qsort_by_w_h_n(const void *a, const void *b);
	static int qsort_by_n_w_h(const void *a, const void *b);
	static BOOL CALLBACK MonitorProc(HMONITOR hMonitor, HDC hdcMonitor, LPRECT lprcMonitor, LPARAM dwData);
	//@}
	/// @name screen private members
	//@{
	// screen settings
	HWND main_window_handle;
	HINSTANCE instance_handle;
	DWORD dwStyle;

	/// vm draw area
	int screen_width, screen_height;
	int screen_width_aspect, screen_height_aspect;
	/// vm draw area (not used)
	int window_width, window_height;
	/// minimum window client size
	int min_window_width, min_window_height;
	/// display size (window client size)
	int display_width, display_height;
	/// desktop size on the monitor
	int desktop_width, desktop_height, desktop_bpp;
	bool screen_size_changed;

	HDC hdcDibSource;
	scrntype* lpBmpSource;
	LPBITMAPINFO lpDibSource;

	RectWH rec_video_size[2];
	HDC hdcDibRecVideo;
	HBITMAP hBmpRecVideo, hOldBmpRecVideo;
	LPBYTE lpBufRecVideo;
	scrntype* lpBmpRecVideo;
	LPBITMAPINFO lpDibRecVideo;
	LPBITMAPINFOHEADER pbmInfoHeader;

	/// vm draw area
	/// if use screen rotation, swap width and height.
	int source_width, source_height;
	int source_width_aspect, source_height_aspect;
	int stretched_width, stretched_height;
	int stretch_pow_x, stretch_pow_y;
	int screen_dest_x, screen_dest_y;
	int display_stretched_width, display_stretched_height;
	int screen_stretched_dest_x, screen_stretched_dest_y;
	bool stretch_screen;

	// update flags
	bool first_draw_screen;
	bool first_invalidate;
	bool self_invalidate;
	bool skip_frame;
	bool skip_frame_pre;

	// screen buffer
	HDC hdcDib;
	HBITMAP hBmp, hOldBmp;
	LPBYTE lpBuf;
	scrntype* lpBmp;
	LPBITMAPINFO lpDib;

#ifdef USE_SCREEN_ROTATE
	// rotate buffer
	HDC hdcDibRotate;
	HBITMAP hBmpRotate, hOldBmpRotate;
	LPBYTE lpBufRotate;
	scrntype* lpBmpRotate;
	LPBITMAPINFO lpDibRotate;
#endif
#ifdef _BML3MK5
	// for message board
	HDC hdcDibMixed;
	HBITMAP hBmpMixed, hOldBmpMixed;
	LPBYTE lpBufMixed;
	scrntype* lpBmpMixed;
	LPBITMAPINFO lpDibMixed;
#endif

	// stretch buffer
	HDC hdcDibStretch1;
	HBITMAP hBmpStretch1, hOldBmpStretch1;
	LPBYTE lpBufStretch1;
	scrntype* lpBmpStretch1;
	LPBITMAPINFO lpDibStretch1;

	HDC hdcDibStretch2;
	HBITMAP hBmpStretch2, hOldBmpStretch2;
	LPBYTE lpBufStretch2;
	scrntype* lpBmpStretch2;
	LPBITMAPINFO lpDibStretch2;

#ifdef USE_DIRECT3D
	PDIRECT3D9			pD3D;
	D3DPRESENT_PARAMETERS	d3dpp;
	D3DDISPLAYMODE			d3ddm;
	PDIRECT3DDEVICE9	pD3Device;
	PDIRECT3DSURFACE9	pD3Dmixsuf;
	PDIRECT3DSURFACE9	pD3Dsource;
	PDIRECT3DSURFACE9	pD3Dorigin;
#ifdef USE_SCREEN_ROTATE
	PDIRECT3DSURFACE9	pD3Drotate;
#endif

	scrntype*			lpD3DBmp;
	bool                lost_d3device;
	bool                enable_direct3d;

	RECT				reD3Dmix;
	RECT				reD3Dsuf;
#endif

	// screen mode
#define DISP_DEVICE_MAX 3
	int disp_device_count;
	int monitor_count;
	struct disp_device_st {
		TCHAR    name[32];
		HMONITOR hMoni;
		RECT     re;
	} disp_device[DISP_DEVICE_MAX];

#define SCREEN_MODE_MAX (ID_SCREEN_FULLSCREEN24 - ID_SCREEN_FULLSCREEN1 + 1)
	int screen_mode_count;
	struct screen_mode_st {
		int disp_no;
		int width;
		int height;
	} screen_mode[SCREEN_MODE_MAX];

#define WINDOW_MODE_MAX (ID_SCREEN_WINDOW8 - ID_SCREEN_WINDOW1 + 1)
	int window_mode_count;
	struct window_mode_st {
		int power;
		int width;
		int height;
	} window_mode[WINDOW_MODE_MAX];
	int window_mode_power;

	bool now_fullscreen;
	int  prev_window_mode;

	// record video
	bool now_rec_vid;
	int rec_frames, rec_fps;
	PAVIFILE pAVIFile;
	PAVISTREAM pAVIStream;
	PAVISTREAM pAVICompressed;
	AVICOMPRESSOPTIONS opts;
	//@}

	// ----------------------------------------
	// sound
	// ----------------------------------------
	/// @name sound private methods
	//@{
	void initialize_sound(int rate, int samples, int latency);
	void initialize_sound(int rate, int samples);
	void initialize_sound();
	void release_sound();
	void update_sound(int* extra_frames);
	//@}
	/// @name sound private members
	//@{
	int sound_rate, sound_samples;
	int sound_latency_half;
	bool sound_ok, sound_started, now_mute;

	// direct sound
	LPDIRECTSOUND lpds;
	LPDIRECTSOUNDBUFFER lpdsb, lpdsp;
	bool first_half;
	DWORD sound_prev_time;

	// record sound
	typedef struct {
		DWORD dwRIFF;
		DWORD dwFileSize;
		DWORD dwWAVE;
		DWORD dwfmt_;
		DWORD dwFormatSize;
		WORD wFormatTag;
		WORD wChannels;
		DWORD dwSamplesPerSec;
		DWORD dwAvgBytesPerSec;
		WORD wBlockAlign;
		WORD wBitsPerSample;
		DWORD dwdata;
		DWORD dwDataLength;
	} wavheader_t;
	FILEIO* rec;
	int rec_bytes;
	bool now_rec_snd;
	//@}

	// ----------------------------------------
	// floppy disk
	// ----------------------------------------
#ifdef USE_FD1
	/// @name floppy disk private methods
	//@{
	void initialize_disk_insert();
	void update_disk_insert();
	//@}
	/// @name floppy disk private members
	//@{
	typedef struct {
		_TCHAR path[_MAX_PATH];
		int offset;
		int wait_count;
		uint32 flags;
	} disk_insert_t;
	disk_insert_t disk_insert[8];
	//@}
#endif

	// ----------------------------------------
	// media
	// ----------------------------------------
#ifdef USE_MEDIA
	/// @name media private methods
	//@{
	void initialize_media();
	void release_media();
	//@}
	/// @name media private members
	//@{
	_TCHAR media_path[MEDIA_MAX][_MAX_PATH];
	int media_cnt;
	bool media_playing;
	//@}
#endif

	// ----------------------------------------
	// timer
	// ----------------------------------------
	/// @name timer private methods
	//@{
	void update_timer();
	//@}
	/// @name timer private members
	//@{
	SYSTEMTIME sTime;
	//@}

	// ----------------------------------------
	// socket
	// ----------------------------------------
#ifdef USE_SOCKET
	/// @name socket private methods
	//@{
	void initialize_socket();
	void release_socket();
	void update_socket();
	//@}
	/// @name socket private members
	//@{
	SOCKET soc[SOCKET_MAX];
	bool is_tcp[SOCKET_MAX];
	struct sockaddr_in udpaddr[SOCKET_MAX];
	int socket_delay[SOCKET_MAX];
	char *recv_buffer[SOCKET_MAX];
	int recv_r_ptr[SOCKET_MAX], recv_w_ptr[SOCKET_MAX];
	//@}
#endif

	// ----------------------------------------
	// misc
	// ----------------------------------------
	/// @name misc private methods
	//@{
	void open_logfile();
	void close_logfile();
#ifdef _BML3MK5
	int update_led();
#endif
	//@}
	/// @name misc private members
	//@{
#ifdef _LOG_FILE
	FILE* fplog;
#endif
	_TCHAR app_path[_MAX_PATH];

	bool vm_pause,vm_syspause,vm_usrpause;

	LedBox *ledbox;
	MsgBoard *msgboard;
	ConfigBox *configbox;
	VolumeBox *volumebox;
	//@}

public:
	KeybindBox *keybindbox;

//
//
//
public:
	// ----------------------------------------
	// initialize
	// ----------------------------------------
	/// @name initialize
	//@{
	EMU();
	~EMU();
	void initialize(HWND hwnd, HINSTANCE hinst);
	void release();
	//@}

	// ----------------------------------------
	// misc
	// ----------------------------------------
	/// @name misc
	//@{
	void    application_path(_TCHAR* path);
	_TCHAR *application_path();

	bool get_pause(int idx);
	void set_pause(int idx, bool val);
	bool *get_pause_ptr(void);
	//@}

	// ----------------------------------------
	// for windows
	// ----------------------------------------
	/// @name drive virtual machine
	//@{
	int frame_interval();
	int run(int split_num);
	void update();
	bool now_skip();
	//@}
	// user interface
	/// @name control menu for ui
	//@{
	void update_config();
	void reset();
#ifdef USE_SPECIAL_RESET
	void special_reset();
	void warm_reset();
#endif
	void change_dipswitch(int num);
#ifdef USE_POWER_OFF
	void notify_power_off();
#endif
	void change_cpu_power(int num);
#ifdef _BML3MK5
	void change_sync_irq();
#endif
#ifdef USE_CART
	void open_cart(_TCHAR* file_path);
	void close_cart();
#endif
#ifdef USE_AUTO_KEY
	void start_auto_key();
	void start_auto_key(_TCHAR *path);
	void stop_auto_key();
	bool now_auto_key() {
		return (autokey_phase != 0);
	}
#endif
#ifdef _BML3MK5
	void save_state(_TCHAR* filename);
	void load_state(_TCHAR* filename);
#endif
	//@}
#ifdef USE_FD1
	/// @name floppy disk menu for ui
	//@{
	bool open_disk(int drv, _TCHAR* file_path, int offset, uint32 flags);
	bool open_disk(int drv, _TCHAR* file_path, int offset);
	void close_disk(int drv);
	int  change_disk(int drv);
	int  get_disk_side(int drv);
	//@}
#endif
#ifdef USE_QUICKDISK
	void open_quickdisk(_TCHAR* file_path);
	void close_quickdisk();
#endif
#ifdef USE_DATAREC
	/// @name tape menu for ui
	//@{
	bool play_datarec(_TCHAR* file_path);
	bool rec_datarec(_TCHAR* file_path);
	void close_datarec();
	void rewind_datarec();
	void fast_forward_datarec();
	void realmode_datarec();
	//@}
#endif
#ifdef USE_DATAREC_BUTTON
	void push_play();
	void push_stop();
#endif
#ifdef USE_MEDIA
	void open_media(_TCHAR* file_path);
	void close_media();
#endif
#ifdef USE_BINARY_FILE1
	void load_binary(int drv, _TCHAR* file_path);
	void save_binary(int drv, _TCHAR* file_path);
#endif
	/// @name screen menu for ui
	//@{
	void change_screen_mode(HWND hWnd, int mode);
	void change_stretch_screen();
	void change_cutout_screen();
	int get_window_mode_count() {
		return window_mode_count;
	}
	int get_screen_mode_count() {
		return screen_mode_count;
	}
	int is_fullscreen() {
		return now_fullscreen;
	}
	void get_window_mode_str(int num, _TCHAR *str);
	void get_screen_mode_str(int num, _TCHAR *str);
	void get_screen_record_size_str(int num, _TCHAR *str);
	void capture_screen();
	void start_rec_video(int fps, bool show_dialog);
	void stop_rec_video();
	void restart_rec_video();
	void resize_rec_video(int new_size);
	bool now_rec_video() {
		return now_rec_vid;
	}
#ifdef USE_SCANLINE
	void change_screen_scanline();
#endif
#ifdef _BML3MK5
	void change_screen_use_direct3d(int num);
	void change_screen_afterimage(int num);
	void change_screen_keepimage(int num);
#endif
#ifdef USE_DIRECT3D
	bool enabled_direct3d() { return enable_direct3d; }
#endif
	//@}
	/// @name sound menu for ui
	//@{
	void create_volume_dialog();
	void start_rec_sound();
	void stop_rec_sound();
	void restart_rec_sound();
	bool now_rec_sound() {
		return now_rec_snd;
	}
	bool *get_now_rec_sound_ptr() {
		return &now_rec_snd;
	}
	void set_volume(int volume);
	//@}
#ifdef USE_PRINTER
	/// @name printer menu for ui
	//@{
	void save_printer(int drv, _TCHAR* filename);
	void clear_printer(int drv);
	int  get_printer_buffer_size(int drv);
	uint8* get_printer_buffer(int drv);
//	void send_printer_data(uint8 *datas, int len);
	void enable_printer_direct(int drv);
	bool print_printer(int drv);
	//@}
#endif
#ifdef _BML3MK5
	/// @name comm menu for ui
	//@{
	void enable_comm_server(int drv);
	void enable_comm_connect(int drv);
	//@}
#endif
#ifdef _BML3MK5
	/// @name options menu for ui
	//@{
	void show_led();
	void inside_led();
	void move_led();
	bool is_shown_ledbox(int id);
	void show_message_board();
	bool is_shown_message_board();
	void toggle_use_joypad();
	bool is_enable_joypad();

	void show_config_dialog();
	void show_keybind_dialog();
	//@}
#endif
	//
	/// @name input device procedures for host machine
	//@{
//	void key_down(int code, bool repeat);
	int  key_down(int code, long status);
//	void key_up(int code);
	void key_up(int code, long status = 0);
	void key_lost_focus() {
		lost_focus = true;
	}
#ifdef USE_BUTTON
	void press_button(int num);
#endif

	void enable_mouse();
	void disenable_mouse();
	void toggle_mouse();
	bool get_mouse_enabled() {
		return mouse_enabled;
	}

#ifdef USE_AUTO_KEY
	void parse_auto_key(char *buf, int size, int &prev_code, int &prev_kana);
#endif
	void enable_joystick();
	//@}
	/// @name screen device procedures for host machine
	//@{
	int get_window_width(int mode);
	int get_window_height(int mode);
	void set_display_size(int width, int height, int power, bool window_mode);
	void change_screen_size(int sw, int sh, int swa, int sha, int ww, int wh);
	void draw_screen();
	void mix_screen();
	void update_screen(HDC hdc);
	void skip_screen() {
		skip_frame = true;
	}

	void enum_window_mode(int max_width, int max_height);
	void enum_screen_mode();
	void find_screen_mode();
	void set_window(HWND hWnd, int mode, int cur_width, int cur_height);
	void set_desktop_size(int width, int height, int bpp) {
		desktop_width = width;
		desktop_height = height;
		desktop_bpp = bpp;
	}
	void set_window_style(DWORD style) {
		dwStyle = style;
	}

#ifdef USE_BITMAP
	void reload_bitmap() {
		first_invalidate = true;
	}
#endif
#ifdef USE_DIRECT3D
	PDIRECT3DDEVICE9 get_d3device() { return pD3Device; }
#endif
	//@}
	/// @name sound device procedures for host machine
	//@{
	void mute_sound(bool mute);
	//@}
#ifdef USE_SOCKET
	/// @name socket device procedures for host machine
	//@{
	SOCKET get_socket(int ch) {
		return soc[ch];
	}
	void socket_connected(int ch);
	void socket_disconnected(int ch);
	void socket_writeable(int ch);
	void socket_readable(int ch);
	void socket_accept(int ch);
	void send_data(int ch);
	void recv_data(int ch);
	//@}
#endif

	// ----------------------------------------
	// for virtual machine
	// ----------------------------------------

	// power off
	void power_off() {
		PostMessage(main_window_handle, WM_CLOSE, 0, 0L);
	}

	/// @name input device for vm
	//@{
	uint8* key_buffer() {
		return key_status;
	}
	uint32* joy_buffer() {
		return joy_status;
	}
	int* mouse_buffer() {
		return mouse_status;
	}
	int* get_key_mod_ptr() {
		return &key_mod;	
	}
	void recognized_key(uint8 key_code);
	//@}
	/// @name screen for vm
	//@{
	scrntype* screen_buffer(int y);
	int screen_buffer_offset();
	void get_rgbformat(uint32 *r_mask, uint32 *g_mask, uint32 *b_mask, uint8 *r_shift, uint8 *g_shift, uint8 *b_shift);
	void get_rgbcolor(scrntype pixel, uint8 *r, uint8 *g, uint8 *b);
	scrntype map_rgbcolor(uint8 r, uint8 g, uint8 b);
	bool* get_skip_frame() {
		return &skip_frame;
	}
	//@}
	/// @name sound for vm
	//@{
	void record_rec_sound(uint16 *buffer, int samples);
	void record_rec_sound(int32 *buffer, int samples);
	//@}
	/// @name timer for vm
	//@{
	void get_timer(int time[]);
	//@}
#ifdef USE_MEDIA
	/// @name media for vm
	//@{
	int media_count();
	void play_media(int trk);
	void stop_media();
	//@}
#endif
#ifdef USE_SOCKET
	/// @name socket for vm
	//@{
	bool init_socket_tcp(int ch, bool server = false);
	bool init_socket_udp(int ch, bool server = false);
	bool connect_socket(int ch, uint32 ipaddr, int port, bool server = false);
	bool connect_socket(int ch, const _TCHAR *hostname, int port, bool server = false);
	bool get_ipaddr(const _TCHAR *hostname, int port, uint32 *ipaddr);
	void disconnect_socket(int ch);
	int  get_socket_channel();
	bool listen_socket(int ch);
	void send_data_tcp(int ch);
	void send_data_udp(int ch, uint32 ipaddr, int port);
	void send_data_udp(int ch, const _TCHAR *hostname, int port);
	//@}
#endif
#ifdef _BML3MK5
//	void send_comm_data(uint8 *data, size_t size);
//	void recv_comm_data(uint8 *data, size_t size);
	/// for get_parami method
	enum enumParamId {
		ConfigBoxFddType,
		ConfigBoxWavReverse,
		ConfigBoxWavHalf,
		ConfigBoxWavCorrect,
		ConfigBoxWavCorrectType,
		ConfigBoxWavSampleRate,
		ConfigBoxWavSampleBits,
		ConfigBoxIOPort,		
	};
	/// @name misc for vm
	//@{
	int get_parami(enumParamId id);
	//@}
#endif
	/// @name logging methods
	//@{
	void out_log(int level, const _TCHAR* format, ...);
	void out_logv(int level, const _TCHAR* format, va_list ap);
	/// debug log
	void out_debug(const _TCHAR* format, ...);
	//
	void out_msg(const _TCHAR* msg, bool set, int sec);
	void out_msg(const _TCHAR* msg, bool set = true);
	void out_info(const _TCHAR* msg, bool set, int sec);
	void out_info(const _TCHAR* msg, bool set = true);
	void out_infof(const _TCHAR* format, ...);
	//@}
};

#endif
