/** @file floppy.cpp

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.11.05 -

	@brief [ floppy drive ]
*/

#include "floppy.h"
#include "mc6843.h"
#include "mb8866.h"
#include "../disk.h"

#define EVENT_IRQ			0
#define EVENT_MOTOR_OFF		1
#define EVENT_INDEXHOLE_ON	2
#define EVENT_INDEXHOLE_OFF	3

//#define DELAY_INDEX_HOLE	200000
#define DELAY_INDEX_HOLE	50000

#define DELAY_MOTOR_OFF_3	2000000
#define DELAY_MOTOR_OFF_5	100000
#define DELAY_MOTOR_TIMEOUT	60000000

#define DRIVE_MASK	(MAX_DRIVE - 1)


void FLOPPY::cancel_my_event(int event_no) {
	if(register_id[event_no] != -1) {
		cancel_event(register_id[event_no]);
		register_id[event_no] = -1;
	}
}
void FLOPPY::register_my_event(int event_no, int wait) {
	if(register_id[event_no] != -1) {
		cancel_event(register_id[event_no]);
		register_id[event_no] = -1;
	}
	register_event(this, event_no, wait, false, &register_id[event_no]);
}

void FLOPPY::initialize()
{
	// setup/reset floppy drive
	for(int i = 0; i < MAX_DRIVE; i++) {
		disk[i] = new DISK();
	}
	for(int i = 0; i < MAX_DRIVE; i++) {
		fdc[i].side = 0;
		fdc[i].track = 0;
		fdc[i].index = 0;
		fdc[i].access = false;
	}
	ignore_crc = false;

	sidereg = 0;
	sectorcnt = 0;

	for(int i = 0; i < FLOPPY_MAX_EVENT; i++) {
		register_id[i] = -1;
	}
}

void FLOPPY::reset()
{
	for(int i = 0; i < MAX_DRIVE; i++) {
		fdc[i].track = 0;
		fdc[i].index = 0;
		fdc[i].access = false;
	}

	for(int i = 0; i < FLOPPY_MAX_EVENT; i++) {
		cancel_my_event(i);
	}

	drvreg = 0;
	drvsel = 0;

	index_hole = 0;
	density = 0;

	outreg = 0x81;
	irqflg = false;
	drqflg = false;

	ignore_write = false;

	register_my_event(EVENT_INDEXHOLE_ON, DELAY_INDEX_HOLE);	// index hole
	register_my_event(EVENT_MOTOR_OFF, DELAY_MOTOR_TIMEOUT);	// delay motor off(timeout) (60s)
}

void FLOPPY::release()
{
	// release d88 handler
	for(int i = 0; i < MAX_DRIVE; i++) {
		delete disk[i];
	}
}

void FLOPPY::write_io8(uint32 addr, uint32 data)
{
	set_drive_select(data);

//	emu->out_debug("fddw a:%04x d:%02x"
//		,addr,data);
}

uint32 FLOPPY::read_io8(uint32 addr)
{
	switch(config.fdd_type) {
		case 1:
			return drvsel;
			break;
		case 2:
			return outreg;
			break;
	}
	return 0;
}

void FLOPPY::write_signal(int id, uint32 data, uint32 mask)
{
	uint8 noutreg = outreg;

	switch (id) {
		case SIG_FLOPPY_ACCESS:
			fdc[drvreg].access = (data & 1) ? true : false;
			break;
		case SIG_FLOPPY_WRITE:
			if(!ignore_write && fdc[drvreg].index < disk[drvreg]->sector_size) {
				disk[drvreg]->sector[fdc[drvreg].index] = (data & mask);
				fdc[drvreg].index++;
			}
			break;
		case SIG_FLOPPY_WRITE_TRACK:
			if(!ignore_write && fdc[drvreg].index < disk[drvreg]->track_size) {
				disk[drvreg]->track[fdc[drvreg].index] = (data & mask);
				fdc[drvreg].index++;
			}
			break;
		case SIG_FLOPPY_WRITEDELETE:
			if (!ignore_write) {
				disk[drvreg]->deleted = (data & 1) ? 0x10 : 0;
			}
			break;
		case SIG_FLOPPY_STEP:
			if (data < 0x80) fdc[drvreg].track++;
			else if (data > 0x80) fdc[drvreg].track--;

			if (fdc[drvreg].track < 0) fdc[drvreg].track = 0;
			if (fdc[drvreg].track > 255) fdc[drvreg].track = 255;
			break;
		case SIG_FLOPPY_TRACK_SIZE:
			disk[drvreg]->track_size = (data & mask);
			fdc[drvreg].index = 0;
			break;
		case SIG_FLOPPY_CURRENTTRACK:
			fdc[drvreg].track = (data & mask);
			break;
		case SIG_FLOPPY_HEADLOAD:
			break;
		case SIG_CPU_RESET:
			now_reset = (data & mask) ? true : false;
			reset();
			if (now_reset) {
				set_drive_select(0);
			}
			break;
	}

	if (config.fdd_type == 2) {
		// for 5inch mini floppy
		switch (id) {
			case SIG_FLOPPY_IRQ:
				irqflg = (data & mask) ? true : false;
				if (irqflg) {
					noutreg &= 0xfe;
				} else {
					noutreg |= 0x01;
				}
				break;
			case SIG_FLOPPY_DRQ:
				drqflg = (data & mask) ? true : false;
				if (drqflg) {
					noutreg &= 0x7f;
				} else {
					noutreg |= 0x80;
				}
				break;
		}

		if (noutreg != outreg) {
//			emu->out_debug("fdd sig id:%04x d:%02x -> %02x"
//				,id,outreg,noutreg);
		}

		if ((drvsel & 0x40) == 0) {	// nmi non mask
			// nmi interrupt
			switch (id) {
			case SIG_FLOPPY_IRQ:
//				set_irq(irqflg);
				register_my_event(EVENT_IRQ, 20);	// delay
				break;
	//		case SIG_FLOPPY_DRQ:
	//			set_drq(drqflg);
	//			break;
			}
		}
		outreg = noutreg;
	}
}

uint32 FLOPPY::read_signal(int id)
{
	uint32 data = 0;

	switch (id) {
		case SIG_FLOPPY_READ_ID:
			if(fdc[drvreg].index < 6) {
				data = disk[drvreg]->id[fdc[drvreg].index];
				fdc[drvreg].index++;
			}
			break;
		case SIG_FLOPPY_READ:
			if(fdc[drvreg].index < disk[drvreg]->sector_size) {
				data = disk[drvreg]->sector[fdc[drvreg].index];
				fdc[drvreg].index++;
			}
			break;
		case SIG_FLOPPY_READ_TRACK:
			if(fdc[drvreg].index < disk[drvreg]->track_size) {
				data = disk[drvreg]->track[fdc[drvreg].index];
				fdc[drvreg].index++;
			}
			break;
		case SIG_FLOPPY_WRITEPROTECT:
			data = (disk[drvreg]->write_protected || ignore_write) ? 1 : 0;
			break;
		case SIG_FLOPPY_READY:
			data = disk[drvreg]->inserted ? 1 : 0;
			break;
		case SIG_FLOPPY_TRACK0:
			data = (fdc[drvreg].track == 0) ? 1 : 0;
			break;
		case SIG_FLOPPY_INDEX:
			data = index_hole;
			break;
		case SIG_FLOPPY_DELETED:
			data = disk[drvreg]->deleted ? 1 : 0;
			break;
		case SIG_FLOPPY_CURRENTTRACK:
			data = fdc[drvreg].track;
			break;
		case SIG_FLOPPY_SECTOR_NUM:
			data = disk[drvreg]->sector_num;
			break;
		case SIG_FLOPPY_SECTOR_SIZE:
			data = disk[drvreg]->sector_size;
			break;
		case SIG_FLOPPY_TRACK_SIZE:
			data = disk[drvreg]->track_size;
			break;
	}

	return data;
}

// ----------------------------------------------------------------------------
// event handler
// ----------------------------------------------------------------------------

void FLOPPY::event_frame()
{
}

void FLOPPY::event_callback(int event_id, int err)
{
	int event_no = event_id;

	register_id[event_no] = -1;

	switch(event_no) {
		case EVENT_MOTOR_OFF:
			// motor off
			if (config.fdd_type == 1) {
				drvsel &= 0x7f;
			} else if (config.fdd_type == 2) {
				drvsel &= 0xf7;
//				d_fdc5->write_signal(SIG_MB8866_MOTOR, drvsel, 0x08);
			}
			ignore_write = false;
			break;
		case EVENT_INDEXHOLE_ON:
			// index hole on
			index_hole = 1;
			register_my_event(EVENT_INDEXHOLE_OFF, 100);	// index hole off
			break;
		case EVENT_INDEXHOLE_OFF:
			// index hole off
			index_hole = 0;
			register_my_event(EVENT_INDEXHOLE_ON, DELAY_INDEX_HOLE - 100);	// index hole on
			break;
		case EVENT_IRQ:
			cancel_my_event(EVENT_IRQ);
			set_irq(irqflg);
			break;
	}

}

// ----------------------------------------------------------------------------
void FLOPPY::set_drive_select(uint8 data)
{
	uint8 ndrvreg = 0;

	if (now_reset) {
		if (config.fdd_type == 2) {
			data = 0;
		} else if (config.fdd_type == 1) {
			data = 0xff;
		}
	}
	// drive number
	if (config.fdd_type == 2) {
		ndrvreg = (data & DRIVE_MASK);

//		d_fdc5->write_signal(SIG_MB8866_DRIVEREG, ndrvreg, DRIVE_MASK);
	} else if (config.fdd_type == 1) {
		switch(data & 0x0f) {
			case 2:
				ndrvreg = 1;
				break;
			case 4:
				ndrvreg = 2;
				break;
			case 8:
				ndrvreg = 3;
				break;
			default:
				ndrvreg = 0;
				break;
		}
		ndrvreg &= DRIVE_MASK;
		if (drvreg != ndrvreg) {
			// update status
			d_fdc3->write_signal(SIG_MC6843_UPDATESTATUS, 0, 0);
		}
	}
	drvreg = ndrvreg;

	if (config.fdd_type == 2) {
		// side select
		set_disk_side(drvreg, (data & 0x10) ? 1 : 0);
//		d_fdc5->write_signal(SIG_MB8866_SIDEREG, data, 0x10);
		// density
		density = (data & 0x20) ? 1 : 0;

		// motor
		if ((data & 0x08) == 0 && (drvsel & 0x08) != 0) {
			// motor on -> off
			register_my_event(EVENT_MOTOR_OFF, DELAY_MOTOR_OFF_5);	// delay motor off (2s)
			data |= 0x08;
			drvsel = data & 0xff;
		} else if ((data & 0x08) != 0) {
			// motor on
			register_my_event(EVENT_MOTOR_OFF, DELAY_MOTOR_TIMEOUT);	// delay motor off(timeout) (60s)
			drvsel = data & 0xff;
		} else {
			// motor off
			drvsel = data & 0xf7;
		}
//		d_fdc5->write_signal(SIG_MB8866_MOTOR, drvsel, 0x08);

	} else if (config.fdd_type == 1) {
		// motor
		if ((data & 0x80) == 0 && (drvsel & 0x80) != 0) {
			// motor on -> off
			register_my_event(EVENT_MOTOR_OFF, DELAY_MOTOR_OFF_3);	// delay motor off (2s)
			data |= 0x80;
			drvsel = data & 0xff;
		} else if ((data & 0x80) != 0) {
			// motor on
			register_my_event(EVENT_MOTOR_OFF, DELAY_MOTOR_TIMEOUT);	// delay motor off(timeout) (60s)
			drvsel = data & 0xff;
		} else {
			// motor off
			drvsel = data & 0x7f;
		}

	}
}

// ----------------------------------------------------------------------------
// media handler
// ----------------------------------------------------------------------------
bool FLOPPY::search_track()
{
	return disk[drvreg]->get_track(fdc[drvreg].track, fdc[drvreg].side);
}

bool FLOPPY::verify_track()
{
	// verify track number
	for(int i = 0; i < disk[drvreg]->sector_num; i++) {
		if(disk[drvreg]->verify[i] != fdc[drvreg].track) {
			return false;
		}
	}
	return true;
}

int FLOPPY::search_sector()
{
	if (!disk[drvreg]->get_sector(fdc[drvreg].track, fdc[drvreg].side, sectorcnt)) {
		return 1;	// SECTOR NOT FOUND
	}
	sectorcnt++;

	fdc[drvreg].index = 0;

	if (disk[drvreg]->status && !ignore_crc) {
		return 2;	// CRC ERROR
	}

	return 0;
}

int FLOPPY::search_sector(int sect)
{
	return search_sector(sect, false, 0);
}

int FLOPPY::search_sector(int sect, bool compare_side, int side)
{
	int status = 0;
	int sector_num = disk[drvreg]->sector_num;
	sectorcnt = sect - 1;
	if(sectorcnt >= sector_num) {
		sectorcnt = 0;
	}
	int index = sectorcnt - 1;
	for(int i = 0; i < sector_num; i++) {
		// get sector
		index++;
		if(index >= sector_num) {
			index = 0;
		}
		if (!disk[drvreg]->get_sector(fdc[drvreg].track, fdc[drvreg].side, index)) {
			continue;
		}

		// check sector in id field
		if(disk[drvreg]->id[2] != sect) {
			continue;
		}
		// check side in id field
		if(compare_side && disk[drvreg]->id[1] != side) {
			continue;
		}

		// check density
//		if(disk[drvreg]->density) {
//			continue;
//		}

		// sector found
		sectorcnt = index;

		fdc[drvreg].index = 0;

		if (disk[drvreg]->status && !ignore_crc) {
			status |= 2;	// CRC ERROR
		}

		return status;
	}
	return 1;	// SECTOR NOT FOUND
}

bool FLOPPY::make_track()
{
	return disk[drvreg]->make_track(fdc[drvreg].track, fdc[drvreg].side);
}

bool FLOPPY::parse_track()
{
	return disk[drvreg]->parse_track(fdc[drvreg].track, fdc[drvreg].side, density);
}

// ----------------------------------------------------------------------------
// irq / drq
// ----------------------------------------------------------------------------
void FLOPPY::set_irq(bool val)
{
	write_signals(&outputs_irq, val ? 0xffffffff : 0);
}

void FLOPPY::set_drq(bool val)
{
	write_signals(&outputs_drq, val ? 0xffffffff : 0);
}

// ----------------------------------------------------------------------------
// user interface
// ----------------------------------------------------------------------------
bool FLOPPY::open_disk(int drv, _TCHAR path[], int offset, uint32 flags)
{
	if(drv < MAX_DRIVE) {
		return disk[drv]->open(path, offset, flags);
	} else {
		return false;
	}
}

void FLOPPY::close_disk(int drv)
{
	if(drv < MAX_DRIVE) {
		disk[drv]->close();
		set_disk_side(drv, 0);
	}
}

int FLOPPY::change_disk(int drv)
{
	if(drv < MAX_DRIVE) {
		set_disk_side(drv, 1 - fdc[drv].side);
		return fdc[drv].side;
	} else {
		return 0;
	}
}

void FLOPPY::set_disk_side(int drv, int side)
{
	fdc[drv].side = side & 1;
	if (fdc[drv].side) {
		sidereg |= (1 << drv);
	} else {
		sidereg &= ~(1 << drv);
	}
}

int FLOPPY::get_disk_side(int drv)
{
	return fdc[drv].side;
}

bool FLOPPY::disk_inserted(int drv)
{
	if(drv < MAX_DRIVE) {
		return disk[drv]->inserted;
	}
	return false;
}

void FLOPPY::set_drive_type(int drv, uint8 type)
{
	if(drv < MAX_DRIVE) {
		disk[drv]->drive_type = type;
	}
}

uint8 FLOPPY::get_drive_type(int drv)
{
	if(drv < MAX_DRIVE) {
		return disk[drv]->drive_type;
	}
	return DRIVE_TYPE_UNK;
}

uint8 FLOPPY::fdc_status()
{
	return 0;
}

// ----------------------------------------------------------------------------
uint8 FLOPPY::get_drive_select()
{
	uint8 data = 0;

	if (config.fdd_type == 2) {
		if (drvsel & 0x08) {
			data = (1 << (drvsel & 0x03)) | (0xf0);
		}
	} else if (config.fdd_type == 1) {
		if (drvsel & 0x80) {
			data = (drvsel & 0x0f) | ((sidereg & 0x0f) << 4);
		}
	}
	return data;
}

// ----------------------------------------------------------------------------
void FLOPPY::save_state(FILEIO *fp)
{
	struct vm_state_st vm_state;

	//
	vm_state_ident.version = 1;
	vm_state_ident.size = sizeof(vm_state_ident) + sizeof(vm_state);

	// copy values
	memset(&vm_state, 0, sizeof(vm_state));
	for(int i=0; i<FLOPPY_MAX_EVENT; i++) {
		vm_state.register_id[i] = register_id[i];
	}
	vm_state.drvreg  = drvreg;
	vm_state.drvsel  = drvsel;
	vm_state.sidereg = sidereg;
	vm_state.outreg  = outreg;
	vm_state.irqflg  = irqflg ? 1 : 0;
	vm_state.drqflg  = drqflg ? 1 : 0;

	fp->Fwrite(&vm_state_ident, sizeof(vm_state_ident), 1);
	fp->Fwrite(&vm_state, sizeof(vm_state), 1);
}

void FLOPPY::load_state(FILEIO *fp)
{
	vm_state_ident_t vm_state_i;
	struct vm_state_st vm_state;

	READ_STATE_CHUNK(fp, vm_state_i, vm_state);

	// copy values
	for(int i=0; i<FLOPPY_MAX_EVENT; i++) {
		register_id[i] = vm_state.register_id[i];
	}
	drvreg  = vm_state.drvreg;
	drvsel  = vm_state.drvsel;
	sidereg = vm_state.sidereg;
	outreg  = vm_state.outreg;
	irqflg  = vm_state.irqflg ? true : false;
	drqflg  = vm_state.drqflg ? true : false;

	// cannot write to disk when load state from file.
	sectorcnt = 0;
	ignore_write = true;
}
