/** @file win32_input.cpp

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.08.18 -

	@note
	Modified for BML3MK5 by Sasaji at 2011.06.17

	@brief [ win32 input ]
*/

#include "emu.h"
#include "vm/vm.h"
#include "fifo.h"
#include "fileio.h"

#define KEY_KEEP_FRAMES 3

void EMU::initialize_input()
{
	// initialize status
	memset(key_status, 0, sizeof(key_status));
	memset(joy_status, 0, sizeof(joy_status));
	memset(mouse_status, 0, sizeof(mouse_status));

	// initialize joysticks
	use_joystick = (config.show_dlg & 4) ? true : false;
	initialize_joystick();

	// mouse emulation is disenabled
	mouse_enabled = false;

#ifdef USE_SHIFT_NUMPAD_KEY
	// initialize shift+numpad conversion
	memset(key_converted, 0, sizeof(key_converted));
	key_shift_pressed = key_shift_released = false;
#endif
#ifdef USE_AUTO_KEY
	// initialize autokey
	autokey_buffer = new FIFO(65536);
	autokey_buffer->clear();
	autokey_phase = autokey_shift = 0;
#ifdef _BML3MK5
	autokey_graph = 0;
#endif
#define AUTO_KEY_KANA	0xf2
#define AUTO_KEY_GRAPH	VK_NONCONVERT
#endif
	lost_focus = false;
	key_mod = 0;
}

void EMU::initialize_joystick()
{
#ifdef USE_JOYSTICK
	memset(joy_enabled, 0, sizeof(joy_enabled));
	memset(joyid_map, 0xff, sizeof(joyid_map));
	reset_joystick();

	memset(&dummy_key, 0, sizeof(dummy_key));
	dummy_key.type = INPUT_KEYBOARD;
	dummy_key.ki.wVk = 0x8f;
	dummy_key.ki.wScan = MapVirtualKey(0x8f, 0);
	dummy_key.ki.dwFlags = KEYEVENTF_EXTENDEDKEY;

	send_dummy_key = 0;
#endif
}

void EMU::reset_joystick()
{
#ifdef USE_JOYSTICK
	if (!use_joystick) return;

	int i = 0;
	joy_num = joyGetNumDevs();
	for(int joyid = 0; joyid < joy_num && joyid < 16 && i < 2; joyid++) {
		JOYINFO joyinfo;
		if(joyGetPos(joyid, &joyinfo) == JOYERR_NOERROR) {
			joyid_map[i] = joyid;

			JOYCAPS joycaps;
			joyGetDevCaps(joyid, &joycaps, sizeof(JOYCAPS));
			uint32 x = (joycaps.wXmin + joycaps.wXmax) / 2;
			joy_xmin[i] = (joycaps.wXmin + x) / 2;
			joy_xmax[i] = (joycaps.wXmax + x) / 2;
			uint32 y = (joycaps.wYmin + joycaps.wYmax) / 2;
			joy_ymin[i] = (joycaps.wYmin + y) / 2;
			joy_ymax[i] = (joycaps.wYmax + y) / 2;

			joy_enabled[i] = true;

			i++;
		}
	}
#endif
}

void EMU::release_input()
{
	// release mouse
	if(mouse_enabled) {
		disenable_mouse();
	}

#ifdef USE_AUTO_KEY
	// release autokey buffer
	if(autokey_buffer) {
		autokey_buffer->release();
		delete autokey_buffer;
	}
#endif
}

void EMU::release_joystick()
{
#ifdef USE_JOYSTICK
	// release joystick
	for(int i = 0; i < 2; i++) {
		joy_enabled[i] = false;
	}
	memset(joy_status, 0, sizeof(joy_status));
	memset(joyid_map, 0xff, sizeof(joyid_map));
#endif
}

void EMU::update_input()
{
#ifdef USE_SHIFT_NUMPAD_KEY
	// update numpad key status
	if(key_shift_pressed && !key_shift_released) {
		if(key_status[VK_SHIFT] == 0) {
			// shift key is newly pressed
			key_status[VK_SHIFT] = 0x80;
#ifdef NOTIFY_KEY_DOWN
			vm->key_down(VK_SHIFT, false);
#endif
		}
	}
	else if(!key_shift_pressed && key_shift_released) {
		if(key_status[VK_SHIFT] != 0) {
			// shift key is newly released
			key_status[VK_SHIFT] = 0;
#ifdef NOTIFY_KEY_DOWN
			vm->key_up(VK_SHIFT);
#endif
			// check l/r shift
			if(!(GetAsyncKeyState(VK_LSHIFT) & 0x8000)) key_status[VK_LSHIFT] &= 0x7f;
			if(!(GetAsyncKeyState(VK_RSHIFT) & 0x8000)) key_status[VK_RSHIFT] &= 0x7f;
		}
	}
	key_shift_pressed = key_shift_released = false;
#endif

	// release keys
#ifdef USE_AUTO_KEY
	if(lost_focus && autokey_phase == 0) {
#else
	if(lost_focus) {
#endif
		// we lost key focus so release all pressed keys
		for(int i = 0; i < 256; i++) {
			if(key_status[i] & 0x80) {
				key_status[i] &= 0x7f;
#ifdef NOTIFY_KEY_DOWN
				if(!key_status[i]) {
					vm->key_up(i);
				}
#endif
			}
		}
	}
	else {
		for(int i = 0; i < 256; i++) {
			if(key_status[i] & 0x7f) {
				key_status[i] = (key_status[i] & 0x80) | ((key_status[i] & 0x7f) - 1);
#ifdef NOTIFY_KEY_DOWN
				if(!key_status[i]) {
					vm->key_up(i);
				}
#endif
			}
		}
	}
	lost_focus = false;

#ifdef USE_JOYSTICK
	update_joystick();
#endif
#ifdef USE_KEY_TO_JOY
	// emulate joystick #1 with keyboard
	if(key_status[0x26]) joy_status[0] |= 0x01;	// up
	if(key_status[0x28]) joy_status[0] |= 0x02;	// down
	if(key_status[0x25]) joy_status[0] |= 0x04;	// left
	if(key_status[0x27]) joy_status[0] |= 0x08;	// right
#endif

	// update mouse status
	memset(mouse_status, 0, sizeof(mouse_status));
	if(mouse_enabled) {
		// get current status
		POINT pt;
		GetCursorPos(&pt);
		ScreenToClient(main_window_handle, &pt);
		mouse_status[0]  = pt.x - display_width / 2;
		mouse_status[1]  = pt.y - display_height / 2;
		mouse_status[2]  = (GetAsyncKeyState(VK_LBUTTON) & 0x8000) ? 1 : 0;
		mouse_status[2] |= (GetAsyncKeyState(VK_RBUTTON) & 0x8000) ? 2 : 0;
		mouse_status[2] |= (GetAsyncKeyState(VK_MBUTTON) & 0x8000) ? 4 : 0;
		// move mouse cursor to the center of window
		if(!(mouse_status[0] == 0 && mouse_status[1] == 0)) {
			pt.x = display_width / 2;
			pt.y = display_height / 2;
			ClientToScreen(main_window_handle, &pt);
			SetCursorPos(pt.x, pt.y);
		}
	}
#ifdef USE_LIGHTPEN
	if(true) {
		// get current status
		POINT pt;
		GetCursorPos(&pt);
		ScreenToClient(main_window_handle, &pt);
		// adjust point
		pt.x = screen_width * (pt.x - screen_dest_x) / stretched_width;
		pt.y = screen_height * (pt.y - screen_dest_y) / stretched_height;

		mouse_status[0] = pt.x;
		mouse_status[1] = pt.y;
		mouse_status[2] = (GetAsyncKeyState(VK_LBUTTON) & 0x8000 ? 1 : 0) | (GetAsyncKeyState(VK_RBUTTON) & 0x8000 ? 2 : 0);
	}
#endif

#ifdef USE_AUTO_KEY
	// auto key
	switch(autokey_phase) {
	case 1:
		if(autokey_buffer && !autokey_buffer->empty()) {
			// update shift key status
			int shift = autokey_buffer->read_not_remove(0) & 0x100;
			if(shift && !autokey_shift) {
//				key_down(VK_SHIFT, false);
				key_down(VK_SHIFT, 0);
			}
			autokey_shift = shift;
#ifdef _BML3MK5
			int graph = autokey_buffer->read_not_remove(0) & 0x200;
			if(graph && !autokey_graph) {
//				key_down(AUTO_KEY_GRAPH, false);
				key_down(AUTO_KEY_GRAPH, 0);
			}
			autokey_graph = graph;
#endif
		}
		autokey_phase++;
		break;
	case 3:
		if(autokey_buffer && !autokey_buffer->empty()) {
//			key_down(autokey_buffer->read_not_remove(0) & 0xff, false);
			autokey_code = autokey_buffer->read_not_remove(0) & 0xff;
			autokey_code = key_down(autokey_code, 0);
		}
		out_debug(_T("autokeydown: phase:%d auto_key:%x"),autokey_phase,autokey_code);
		autokey_phase++;
		break;
	case 4:
		if(!autokey_buffer || autokey_buffer->empty()) {
			autokey_phase++;
		}
		// wait response from vm
		break;
	case USE_AUTO_KEY:
		if(autokey_shift) {
			key_up(VK_SHIFT);
			autokey_shift = 0;
		}
#ifdef _BML3MK5
		if(autokey_graph) {
			key_up(AUTO_KEY_GRAPH);
			autokey_graph = 0;
		}
#endif
		if(autokey_buffer && !autokey_buffer->empty()) {
			key_up(autokey_buffer->read_not_remove(0) & 0xff);
		}
		out_debug(_T("autokeyup: phase:%d auto_key:%x"),autokey_phase,autokey_code);
		autokey_code = 0;
		autokey_phase++;
		break;
	case USE_AUTO_KEY + 1:
		if(!autokey_buffer || autokey_buffer->empty()) {
			autokey_phase++;
		}
		// wait response from vm
		break;
	case USE_AUTO_KEY_RELEASE:
		if(autokey_buffer && !autokey_buffer->empty()) {
			// wait enough while vm analyzes one line
			if(autokey_buffer->read() == 0xd) {
				autokey_phase++;
				break;
			}
		}
		// through
	case 30:
		if(autokey_buffer && !autokey_buffer->empty()) {
			autokey_phase = 1;
		}
		else {
			stop_auto_key();
		}
		break;
	default:
		if(autokey_phase) {
			autokey_phase++;
		}
		break;
	}
#endif
}

void EMU::recognized_key(uint8 key_code)
{
	if(autokey_phase == 0) return;
	if(autokey_phase == 4 && autokey_code) {
		out_debug(_T("recognized_key: phase:%d auto_key:%x vm_key:%x"),autokey_phase,autokey_code,key_code);
		if (autokey_code == key_code) autokey_phase++;
	}
	if(autokey_phase == USE_AUTO_KEY + 1 && key_code == 0) {
		out_debug(_T("recognized_key: phase:%d auto_key:%x vm_key:%x"),autokey_phase,autokey_code,key_code);
		autokey_phase++;
	}
}

void EMU::update_joystick()
{
#ifdef USE_JOYSTICK
	if (!use_joystick) return;

	// update joystick status
	memset(joy_status, 0, sizeof(joy_status));
	for(int i = 0; i < 2; i++) {
		JOYINFOEX joyinfoex;
		joyinfoex.dwSize = sizeof(JOYINFOEX);
		joyinfoex.dwFlags = JOY_RETURNALL;
		if(joy_enabled[i] == true && joyid_map[i] != -1) {
			if (joyGetPosEx(joyid_map[i], &joyinfoex) == JOYERR_NOERROR) {
				if(joyinfoex.dwYpos < joy_ymin[i]) joy_status[i] |= 0x01;
				if(joyinfoex.dwYpos > joy_ymax[i]) joy_status[i] |= 0x02;
				if(joyinfoex.dwXpos < joy_xmin[i]) joy_status[i] |= 0x04;
				if(joyinfoex.dwXpos > joy_xmax[i]) joy_status[i] |= 0x08;
				joy_status[i] |= ((joyinfoex.dwButtons & 0x0fffffff) << 4);
			} else {
				joy_enabled[i] = false;
			}
		}
		if (joy_status[i] && send_dummy_key <= 0) {
			// send dummy key event, because delay start time of screen saver.
			SendInput(1, &dummy_key, sizeof(INPUT));
			send_dummy_key = 60;
		} else {
			send_dummy_key -= (send_dummy_key > 0) ? 1 : 0;
		}
	}
#endif
}

#ifdef USE_SHIFT_NUMPAD_KEY
static const int numpad_table[256] = {
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x65, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
//	0x00, 0x69, 0x63, 0x61, 0x67, 0x64, 0x68, 0x66, 0x62, 0x00, 0x00, 0x00, 0x00, 0x60, 0x6e, 0x00,
	0x00, 0x69, 0x63, 0x61, 0x67, 0x64, 0x68, 0x66, 0x62, 0x00, 0x00, 0x00, 0x00, 0x60, 0x00, 0x00,	// remove shift + comma
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};
#endif

#ifdef _BML3MK5
// numpad key convert
// 0x47 - 0x53
static const int scancode2vkey[16] = {
//	0x47,      0x48,      0x49,   0x4a,0x4b,      0x4c,      0x4d,   0x4e,0x4f,      0x50,      0x51,0x52,0x53,0x54,0x55,0x56
	VK_NUMPAD7,VK_NUMPAD8,VK_NUMPAD9,0,VK_NUMPAD4,VK_NUMPAD5,VK_NUMPAD6,0,VK_NUMPAD1,VK_NUMPAD2,VK_NUMPAD3,VK_NUMPAD0,VK_DECIMAL,0,0,0
};
#endif

//void EMU::key_down(int code, bool repeat)
int EMU::key_down(int code, long status)
{
	int  orig_code = code;
	bool keep_frames = false;
	bool repeat = ((HIWORD(status) & 0x4000) != 0);
	int  scan_code = (status & 0x1ff0000) >> 16;

	if(code == VK_SHIFT) {
		if(GetAsyncKeyState(VK_LSHIFT) & 0x8000) key_status[VK_LSHIFT] = 0x80;
		if(GetAsyncKeyState(VK_RSHIFT) & 0x8000) key_status[VK_RSHIFT] = 0x80;
		if(!(key_status[VK_LSHIFT] || key_status[VK_RSHIFT])) key_status[VK_LSHIFT] = 0x80;
	}
	else if(code == VK_CONTROL) {
		if(GetAsyncKeyState(VK_LCONTROL) & 0x8000) key_status[VK_LCONTROL] = 0x80;
		if(GetAsyncKeyState(VK_RCONTROL) & 0x8000) key_status[VK_RCONTROL] = 0x80;
		if(!(key_status[VK_LCONTROL] || key_status[VK_RCONTROL])) key_status[VK_LCONTROL] = 0x80;
	}
	else if(code == VK_MENU) {
		if(GetAsyncKeyState(VK_LMENU) & 0x8000) key_status[VK_LMENU] = 0x80;
		if(GetAsyncKeyState(VK_RMENU) & 0x8000) key_status[VK_RMENU] = 0x80;
		if(!(key_status[VK_LMENU] || key_status[VK_RMENU])) key_status[VK_LMENU] = 0x80;
	}
	else if(code == 0xf0) {
		code = VK_CAPITAL;
		keep_frames = true;
	}
#ifdef _BML3MK5
	else if(code == 0xf1) {
		code = VK_NONCONVERT;
//		keep_frames = true;
	}
#endif
	else if(code == 0xf2) {
		code = VK_KANA;
		keep_frames = true;
	}
	else if(code == 0xf3 || code == 0xf4) {
		code = VK_KANJI;
//		keep_frames = true;
	}

#ifdef _BML3MK5
	// convert numpad keys
	if (scan_code >= 0x47 && scan_code <= 0x53) {
		if (scancode2vkey[scan_code - 0x47] != 0) code = scancode2vkey[scan_code - 0x47];
	}
#endif

#ifdef USE_SHIFT_NUMPAD_KEY
	if(code == VK_SHIFT) {
		key_shift_pressed = true;
		return;
	}
	else if(numpad_table[code] != 0) {
		if(key_shift_pressed || key_shift_released) {
			key_converted[code] = 1;
			key_shift_pressed = true;
			code = numpad_table[code];
		}
	}
#endif
#ifdef DONT_KEEEP_KEY_PRESSED
		if(!(code == VK_SHIFT || code == VK_CONTROL || code == VK_MENU)) {
			key_status[code] = KEY_KEEP_FRAMES;
		}
		else
#endif
	key_status[code] = keep_frames ? (KEY_KEEP_FRAMES * FRAME_SPLIT_NUM) : 0x80;
#ifdef NOTIFY_KEY_DOWN
	if(keep_frames) {
		repeat = false;
	}
	vm->key_down(code, repeat);
#endif
//	out_debug("keydown: %02x->%02x s:%08x repeat:%s",orig_code,code,status,repeat ? "true" : "false");
	return code;
}

//void EMU::key_up(int code)
void EMU::key_up(int code, long status)
{
	int  orig_code = code;
	int  scan_code = (status & 0x1ff0000) >> 16;

	if(code == VK_SHIFT) {
#ifndef USE_SHIFT_NUMPAD_KEY
		if(!(GetAsyncKeyState(VK_LSHIFT) & 0x8000)) key_status[VK_LSHIFT] &= 0x7f;
		if(!(GetAsyncKeyState(VK_RSHIFT) & 0x8000)) key_status[VK_RSHIFT] &= 0x7f;
#endif
	}
	else if(code == VK_CONTROL) {
		if(!(GetAsyncKeyState(VK_LCONTROL) & 0x8000)) key_status[VK_LCONTROL] &= 0x7f;
		if(!(GetAsyncKeyState(VK_RCONTROL) & 0x8000)) key_status[VK_RCONTROL] &= 0x7f;
	}
	else if(code == VK_MENU) {
		if(!(GetAsyncKeyState(VK_LMENU) & 0x8000)) key_status[VK_LMENU] &= 0x7f;
		if(!(GetAsyncKeyState(VK_RMENU) & 0x8000)) key_status[VK_RMENU] &= 0x7f;
	}
	else if(code == 0xf0) {
		code = VK_CAPITAL;
	}
#ifdef _BML3MK5
	else if(code == 0xf1) {
		code = VK_NONCONVERT;
	}
#endif
	else if(code == 0xf2) {
		code = VK_KANA;
	}
	else if(code == 0xf3 || code == 0xf4) {
		code = VK_KANJI;
	}

#ifdef _BML3MK5
	// convert numpad keys
	if (scan_code >= 0x47 && scan_code <= 0x53) {
		if (scancode2vkey[scan_code - 0x47] != 0) code = scancode2vkey[scan_code - 0x47];
	}
#endif

#ifdef USE_SHIFT_NUMPAD_KEY
	if(code == VK_SHIFT) {
		key_shift_pressed = false;
		key_shift_released = true;
		return;
	}
	else if(key_converted[code] != 0) {
		key_converted[code] = 0;
		code = numpad_table[code];
	}
#endif
	if(key_status[code]) {
		key_status[code] &= 0x7f;
#ifdef NOTIFY_KEY_DOWN
		if(!key_status[code]) {
			vm->key_up(code);
		}
#endif
	}
//	out_debug("keyup: %02x->%02x s:%08x",orig_code,code,status);
}

#ifdef USE_BUTTON
void EMU::press_button(int num)
{
	int code = buttons[num].code;

	if(code) {
		key_down(code, false);
		key_status[code] = KEY_KEEP_FRAMES;
	}
	else {
		// code=0: reset virtual machine
		vm->reset();
	}
}
#endif

void EMU::enable_mouse()
{
	// enable mouse emulation
	if(!mouse_enabled) {
		// hide mouse cursor
		ShowCursor(FALSE);
		// move mouse cursor to the center of window
		POINT pt;
		pt.x = display_width / 2;
		pt.y = display_height / 2;
		ClientToScreen(main_window_handle, &pt);
		SetCursorPos(pt.x, pt.y);
	}
	mouse_enabled = true;
}

void EMU::disenable_mouse()
{
	// disenable mouse emulation
	if(mouse_enabled) {
		ShowCursor(TRUE);
	}
	mouse_enabled = false;
}

void EMU::toggle_mouse()
{
	// toggle mouse enable / disenable
	if(mouse_enabled) {
		disenable_mouse();
	}
	else {
		enable_mouse();
	}
}

void EMU::enable_joystick()
{
	// reinitialize
	initialize_joystick();
}

#ifdef USE_AUTO_KEY
static const int autokey_table[256] = {
#ifdef _BML3MK5
	// 0x100: shift
	// 0x200: graph
	// 0x400: alphabet
	// 0x800: ALPHABET
	// 0x1000: katakana
	0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x00d,0x000,0x000,0x00d,0x000,0x000,
	0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
	0x020,0x131,0x132,0x133,0x134,0x135,0x136,0x137,0x138,0x139,0x1ba,0x1bb,0x0bc,0x0bd,0x0be,0x0bf,
	0x030,0x031,0x032,0x033,0x034,0x035,0x036,0x037,0x038,0x039,0x0ba,0x0bb,0x1bc,0x1bd,0x1be,0x1bf,
	0x0c0,0x441,0x442,0x443,0x444,0x445,0x446,0x447,0x448,0x449,0x44a,0x44b,0x44c,0x44d,0x44e,0x44f,
	0x450,0x451,0x452,0x453,0x454,0x455,0x456,0x457,0x458,0x459,0x45a,0x0db,0x0dc,0x0dd,0x0de,0x1e2,
	0x000,0x841,0x842,0x843,0x844,0x845,0x846,0x847,0x848,0x849,0x84a,0x84b,0x84c,0x84d,0x84e,0x84f,
	0x850,0x851,0x852,0x853,0x854,0x855,0x856,0x857,0x858,0x859,0x85a,0x000,0x000,0x000,0x000,0x000,
	// hiragana1
	0x25a,0x258,0x243,0x256,0x242,0x24e,0x24d,0x2bc,0x2be,0x24c,0x24f,0x239,0x238,0x249,0x24b,0x265,
	0x262,0x268,0x266,0x264,0x2bb,0x26d,0x26b,0x2bf,0x267,0x269,0x261,0x263,0x252,0x254,0x246,0x247,
	// katakana
	0x0231,0x11be,0x11db,0x11dd,0x11bc,0x11bf,0x1130,0x1133,0x1145,0x1134,0x1135,0x1136,0x1137,0x1138,0x1139,0x115a,
	0x10dc,0x1033,0x1045,0x1034,0x1035,0x1036,0x1054,0x1047,0x1048,0x10ba,0x1042,0x1058,0x1044,0x1052,0x1050,0x1043,
	0x1051,0x1041,0x105a,0x1057,0x1053,0x1055,0x1049,0x1031,0x10bc,0x104b,0x1046,0x1056,0x1032,0x10de,0x10bd,0x104a,
	0x104e,0x10dd,0x10bf,0x104d,0x1037,0x1038,0x1039,0x104f,0x104c,0x10be,0x10bb,0x10e2,0x1030,0x1059,0x10c0,0x10db,
	// hiragana2
	0x2bd,0x2c0,0x2ba,0x2e2,0x259,0x255,0x248,0x24a,0x251,0x241,0x257,0x253,0x245,0x244,0x26f,0x26a,
	0x260,0x2dc,0x232,0x233,0x234,0x235,0x236,0x237,0x26e,0x230,0x250,0x2de,0x2db,0x2059,0x2dd,0x000,
#else
	// 0x100: shift
	// 0x200: kana
	// 0x400: alphabet
	// 0x800: ALPHABET
	0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x00d,0x000,0x000,0x00d,0x000,0x000,
	0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
	0x020,0x131,0x132,0x133,0x134,0x135,0x136,0x137,0x138,0x139,0x1ba,0x1bb,0x0bc,0x0bd,0x0be,0x0bf,
	0x030,0x031,0x032,0x033,0x034,0x035,0x036,0x037,0x038,0x039,0x0ba,0x0bb,0x1bc,0x1bd,0x1be,0x1bf,
	0x0c0,0x441,0x442,0x443,0x444,0x445,0x446,0x447,0x448,0x449,0x44a,0x44b,0x44c,0x44d,0x44e,0x44f,
	0x450,0x451,0x452,0x453,0x454,0x455,0x456,0x457,0x458,0x459,0x45a,0x0db,0x0dc,0x0dd,0x0de,0x1e2,
	0x1c0,0x841,0x842,0x843,0x844,0x845,0x846,0x847,0x848,0x849,0x84a,0x84b,0x84c,0x84d,0x84e,0x84f,
	0x850,0x851,0x852,0x853,0x854,0x855,0x856,0x857,0x858,0x859,0x85a,0x1db,0x1dc,0x1dd,0x1de,0x000,
	0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
	0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
	// kana -->
	0x000,0x3be,0x3db,0x3dd,0x3bc,0x3bf,0x330,0x333,0x345,0x334,0x335,0x336,0x337,0x338,0x339,0x35a,
	0x2dc,0x233,0x245,0x234,0x235,0x236,0x254,0x247,0x248,0x2ba,0x242,0x258,0x244,0x252,0x250,0x243,
	0x251,0x241,0x25a,0x257,0x253,0x255,0x249,0x231,0x2bc,0x24b,0x246,0x256,0x232,0x2de,0x2bd,0x24a,
	0x24e,0x2dd,0x2bf,0x24d,0x237,0x238,0x239,0x24f,0x24c,0x2be,0x2bb,0x2e2,0x230,0x259,0x2c0,0x2db,
	// <--- kana
	0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
	0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000
#endif
};

void EMU::start_auto_key()
{
	stop_auto_key();

	autokey_code = 0;
	if(OpenClipboard(NULL)) {
		HANDLE hClip = GetClipboardData(CF_TEXT);
		if(hClip) {
			autokey_buffer->clear();
			int prev_code = 0;
			char* buf = (char*)GlobalLock(hClip);
			int size = (int)strlen(buf), prev_kana = 0;
			parse_auto_key(buf, size, prev_code, prev_kana);
			if(prev_kana) {
				autokey_buffer->write(AUTO_KEY_KANA);
#ifdef _BML3MK5
				if (prev_kana == 0x1000) autokey_buffer->write(AUTO_KEY_KANA);
#endif
			}
			GlobalUnlock(hClip);

			autokey_phase = 1;
			autokey_shift = 0;
#ifdef _BML3MK5
			autokey_graph = 0;
#endif
		}
		CloseClipboard();
	}
}

void EMU::start_auto_key(_TCHAR *file_path)
{
	FILEIO *fp;
	char buf[1025];

	stop_auto_key();

	autokey_code = 0;
	fp = new FILEIO();
	int prev_kana = 0;
	if (fp->Fopen(file_path, FILEIO_READ_BINARY)) {
		autokey_buffer->clear();
		int prev_code = 0;
		int size = 0;
		while((size = (int)fp->Fread(buf, sizeof(char), 1024)) > 0) {
			parse_auto_key(buf, size, prev_code, prev_kana);
		}
		fp->Fclose();

		if(prev_kana) {
			autokey_buffer->write(AUTO_KEY_KANA);
#ifdef _BML3MK5
			if (prev_kana == 0x1000) autokey_buffer->write(AUTO_KEY_KANA);
#endif
		}

		autokey_phase = 1;
		autokey_shift = 0;
#ifdef _BML3MK5
		autokey_graph = 0;
#endif
	}
	delete fp;
}

void EMU::parse_auto_key(char *buf, int size, int &prev_code, int &prev_kana)
{
	for(int i = 0; i < size; i++) {
		int code = buf[i] & 0xff;
#ifdef _BML3MK5
		if(prev_code == 0xd && code == 0xa) {
			prev_code = code;
			continue;	// cr-lf
		}
#else
		if((0x81 <= code && code <= 0x9f) || 0xe0 <= code) {
			i++;	// kanji ?
			continue;
		}
		else if(code == 0xa) {
			continue;	// cr-lf
		}
#endif
		if((code = autokey_table[code]) != 0) {
#ifdef _BML3MK5
			int kana = code & 0x3000;
			if(kana - prev_kana == 0x1000) {
				autokey_buffer->write(AUTO_KEY_KANA);
			} else if(kana - prev_kana == 0x2000) {
				autokey_buffer->write(AUTO_KEY_KANA);
				autokey_buffer->write(AUTO_KEY_KANA);
			}
			if(prev_kana - kana == 0x2000) {
				autokey_buffer->write(AUTO_KEY_KANA);
			} else if(prev_kana - kana == 0x1000) {
				autokey_buffer->write(AUTO_KEY_KANA);
				autokey_buffer->write(AUTO_KEY_KANA);
			}
#else
			int kana = code & 0x200;
			if(prev_kana != kana) {
				autokey_buffer->write(AUTO_KEY_KANA);
			}
#endif
			prev_kana = kana;
#if defined(USE_AUTO_KEY_NO_CAPS)
			if((code & 0x100) && !(code & (0x400 | 0x800))) {
#elif defined(USE_AUTO_KEY_CAPS)
			if(code & (0x100 | 0x800)) {
#else
			if(code & (0x100 | 0x400)) {
#endif
				autokey_buffer->write((code & 0xff) | 0x100);
			}
#ifdef _BML3MK5
			else if (code & 0x200) {
				autokey_buffer->write((code & 0xff) | 0x200);
			}
#endif
			else {
				autokey_buffer->write(code & 0xff);
			}
		}
		prev_code =code;
	}
}

void EMU::stop_auto_key()
{
	if(autokey_shift) {
		key_up(VK_SHIFT);
	}
	autokey_phase = autokey_shift = 0;
#ifdef _BML3MK5
	if(autokey_graph) {
		key_up(VK_NONCONVERT);
	}
	autokey_graph = 0;

	memset(key_status, 0, sizeof(key_status));
#endif
}
#endif
