/** @file win32_screen.cpp

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.08.18 -

	@brief [ win32 screen ]
*/

#include "emu.h"
#include "vm/vm.h"
#include "config.h"

static const SIZE window_proposal[] = {
	{640,400},{640,480},{800,600},{1024,768},{0,0}
};

extern void show_menu_bar(HWND hWnd);
extern void hide_menu_bar(HWND hWnd);

void EMU::EMU_SCREEN()
{
	screen_width = SCREEN_WIDTH;
	screen_height = SCREEN_HEIGHT;
	screen_width_aspect = SCREEN_WIDTH_ASPECT;
	screen_height_aspect = SCREEN_HEIGHT_ASPECT;
	window_width = WINDOW_WIDTH;
	window_height = WINDOW_HEIGHT;
	min_window_width = MIN_WINDOW_WIDTH;
	min_window_height = MIN_WINDOW_HEIGHT;
	screen_size_changed = true;
	desktop_width = MIN_WINDOW_WIDTH;
	desktop_height = MIN_WINDOW_HEIGHT;
	desktop_bpp = 24;
	dwStyle = 0;

	source_width = screen_width;
	source_height = screen_height;
	source_width_aspect = source_height_aspect = -1;
	stretch_pow_x = stretch_pow_y = -1;
	display_stretched_width = display_stretched_height = 1;
	stretch_screen = false;

	// screen mode
	disp_device_count = 0;
	monitor_count = 0;
	screen_mode_count = 0;
	window_mode_count = 0;
	window_mode_power = 10;
	now_fullscreen = false;
	prev_window_mode = 0;

	rec_video_size[0].w = MIN_WINDOW_WIDTH;
	rec_video_size[0].h = MIN_WINDOW_HEIGHT;
	rec_video_size[0].x = (MAX_WINDOW_WIDTH - MIN_WINDOW_WIDTH) / 2;
	rec_video_size[0].y = (MAX_WINDOW_HEIGHT - MIN_WINDOW_HEIGHT) / 2;
	rec_video_size[1].w = MAX_WINDOW_WIDTH;
	rec_video_size[1].h = MAX_WINDOW_HEIGHT;
	rec_video_size[1].x = 0;
	rec_video_size[1].y = 0;

#ifdef USE_DIRECT3D
	pD3D = NULL;
	pD3Device = NULL;
	pD3Dmixsuf = NULL;
	pD3Dorigin = NULL;
	pD3Dsource = NULL;
#endif

	// initialize video recording
	now_rec_vid = false;
	pAVIStream = NULL;
	pAVICompressed = NULL;
	pAVIFile = NULL;
	memset(&opts, 0, sizeof(opts));

	// initialize update flags
	first_draw_screen = false;
	first_invalidate = self_invalidate = false;
	skip_frame = false;

}

void EMU::initialize_screen()
{
	// create dib sections
	HDC hdc = GetDC(main_window_handle);
	create_dib_section(hdc, screen_width, screen_height, &hdcDib, &hBmp, &hOldBmp, &lpBuf, &lpBmp, &lpDib);
#ifdef USE_SCREEN_ROTATE
	create_dib_section(hdc, screen_height, screen_width, &hdcDibRotate, &hBmpRotate, &hOldBmpRotate, &lpBufRotate, &lpBmpRotate, &lpDibRotate);
#endif
#ifdef _BML3MK5
	// for message board
	create_dib_section(hdc, source_width, source_height, &hdcDibMixed, &hBmpMixed, &hOldBmpMixed, &lpBufMixed, &lpBmpMixed, &lpDibMixed);
	// for record video
	create_dib_rec_video(hdc, config.screen_video_size);
#endif
	hdcDibStretch1 = hdcDibStretch2 = NULL;
	hBmpStretch1 = hOldBmpStretch1 = hBmpStretch2 = hOldBmpStretch2 = NULL;
	lpBufStretch1 = lpBufStretch2 = NULL;

#ifdef USE_DIRECT3D
	initialize_d3device(main_window_handle);
	create_d3device(main_window_handle);
	create_d3dofflinesurface();
	if (pD3Device != NULL) {
		pD3Device->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0,0,0), 0.0, 0);
	}
#endif
#ifdef _BML3MK5
	if (msgboard) {
		msgboard->InitScreen(hdc, screen_width, screen_height);
	}
	if (ledbox) {
		ledbox->InitScreen(hdc, screen_width, screen_height);
		ledbox->Create(main_window_handle);
		ledbox->Show(config.show_dlg & 9);
	}
#endif
	ReleaseDC(main_window_handle, hdc);
}

#define release_dib_section(hdcdib, hbmp, holdbmp, lpbuf) { \
	if(hdcdib != NULL && holdbmp != NULL) { \
		SelectObject(hdcdib, holdbmp); \
	} \
	if(hbmp != NULL) { \
		DeleteObject(hbmp); \
		hbmp = NULL; \
	} \
	if(lpbuf != NULL) { \
		GlobalFree(lpbuf); \
		lpbuf = NULL; \
	} \
	if(hdcdib != NULL) { \
		DeleteDC(hdcdib); \
		hdcdib = NULL; \
	} \
}

void EMU::release_screen()
{
	// stop video recording
	stop_rec_video();

	// release dib sections
	release_dib_section(hdcDib, hBmp, hOldBmp, lpBuf);
#ifdef USE_SCREEN_ROTATE
	release_dib_section(hdcDibRotate, hBmpRotate, hOldBmpRotate, lpBufRotate);
#endif
#ifdef _BML3MK5
	// for message board
	release_dib_section(hdcDibMixed, hBmpMixed, hOldBmpMixed, lpBufMixed);
	// for record video
	release_dib_section(hdcDibRecVideo, hBmpRecVideo, hOldBmpRecVideo, lpBufRecVideo);
#endif
#ifdef USE_SMOOTH_STRETCH
	release_dib_section(hdcDibStretch1, hBmpStretch1, hOldBmpStretch1, lpBufStretch1);
	release_dib_section(hdcDibStretch2, hBmpStretch2, hOldBmpStretch2, lpBufStretch2);
#endif
#ifdef USE_DIRECT3D
	release_d3device();
	terminate_d3device();
#endif
}

void EMU::create_dib_section(HDC hdc, int width, int height, HDC *hdcDib, HBITMAP *hBmp, HBITMAP *hOldBmp, LPBYTE *lpBuf, scrntype **lpBmp, LPBITMAPINFO *lpDib)
{
	*hdcDib = CreateCompatibleDC(hdc);
	*lpBuf = (LPBYTE)GlobalAlloc(GPTR, sizeof(BITMAPINFO));
	*lpDib = (LPBITMAPINFO)(*lpBuf);
	memset(&(*lpDib)->bmiHeader, 0, sizeof(BITMAPINFOHEADER));
	(*lpDib)->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	(*lpDib)->bmiHeader.biWidth = width;
	(*lpDib)->bmiHeader.biHeight = height;
	(*lpDib)->bmiHeader.biPlanes = 1;
#if defined(_RGB555)
	(*lpDib)->bmiHeader.biBitCount = 16;
	(*lpDib)->bmiHeader.biCompression = BI_RGB;
	(*lpDib)->bmiHeader.biSizeImage = width * height * 2;
#elif defined(_RGB565)
	(*lpDib)->bmiHeader.biBitCount = 16;
	(*lpDib)->bmiHeader.biCompression = BI_BITFIELDS;
	LPDWORD lpBf = (LPDWORD)*lpDib->bmiColors;
	lpBf[0] = 0x1f << 11;
	lpBf[1] = 0x3f << 5;
	lpBf[2] = 0x1f << 0;
	(*lpDib)->bmiHeader.biSizeImage = width * height * 2;
#elif defined(_RGB888)
	(*lpDib)->bmiHeader.biBitCount = 32;
	(*lpDib)->bmiHeader.biCompression = BI_RGB;
	(*lpDib)->bmiHeader.biSizeImage = width * height * 4;
#endif
	(*lpDib)->bmiHeader.biXPelsPerMeter = 0;
	(*lpDib)->bmiHeader.biYPelsPerMeter = 0;
	(*lpDib)->bmiHeader.biClrUsed = 0;
	(*lpDib)->bmiHeader.biClrImportant = 0;
	*hBmp = CreateDIBSection(hdc, *lpDib, DIB_RGB_COLORS, (PVOID*)&(*lpBmp), NULL, 0);
	*hOldBmp = (HBITMAP)SelectObject(*hdcDib, *hBmp);
}

void EMU::get_rgbformat(uint32 *r_mask, uint32 *g_mask, uint32 *b_mask, uint8 *r_shift, uint8 *g_shift, uint8 *b_shift)
{
	*r_mask = 0x00ff0000;
	*g_mask = 0x0000ff00;
	*b_mask = 0x000000ff;
	*r_shift = 16;
	*g_shift = 8;
	*b_shift = 0;
}

void EMU::get_rgbcolor(scrntype pixel, uint8 *r, uint8 *g, uint8 *b)
{
	*r = (pixel & 0x00ff0000) >> 16;
	*g = (pixel & 0x0000ff00) >> 8;
	*b = (pixel & 0x000000ff);
}

scrntype EMU::map_rgbcolor(uint8 r, uint8 g, uint8 b)
{
	return (scrntype)RGB_COLOR(r, g, b);
}

int EMU::get_window_width(int mode)
{
#ifdef USE_SCREEN_ROTATE
	if(config.monitor_type) {
		return window_height + screen_height_aspect * mode;
	}
#endif
	return window_width + screen_width_aspect * mode;
}

int EMU::get_window_height(int mode) {
#ifdef USE_SCREEN_ROTATE
	if(config.monitor_type) {
		return window_width + screen_width_aspect * mode;
	}
#endif
	return window_height + screen_height_aspect * mode;
}

void EMU::set_display_size(int width, int height, int power, bool window_mode)
{
	bool display_size_changed = false;
	bool stretch_changed = false;

	if(width != -1 && (display_width != width || display_height != height)) {
		display_width = width;
		display_height = height;
		display_size_changed = stretch_changed = true;
	}

#ifdef USE_SCREEN_ROTATE
	int prev_source_width = source_width;
	int prev_source_height = source_height;

	if(config.monitor_type) {
		hdcDibSource = hdcDibRotate;
		lpBmpSource = lpBmpRotate;
		lpDibSource = lpDibRotate;

//		lpBmpRecVideo = lpBmpRotate;
//		lpDibRecVideo = lpDibRotate;
//		pbmInfoHeader = &lpDibRotate->bmiHeader;

		stretch_changed |= (source_width != screen_height);
		stretch_changed |= (source_height != screen_width);
		stretch_changed |= (source_width_aspect != screen_height_aspect);
		stretch_changed |= (source_height_aspect != screen_width_aspect);

		source_width = screen_height;
		source_height = screen_width;
		source_width_aspect = screen_height_aspect;
		source_height_aspect = screen_width_aspect;
	}
	else {
#endif
		hdcDibSource = hdcDib;
		lpBmpSource = lpBmp;
		lpDibSource = lpDib;

//		lpBmpRecVideo = lpBmp;
//		lpDibRecVideo = lpDib;
//		pbmInfoHeader = &lpDib->bmiHeader;

		stretch_changed |= (source_width != screen_width);
		stretch_changed |= (source_height != screen_height);
		stretch_changed |= (source_width_aspect != screen_width_aspect);
		stretch_changed |= (source_height_aspect != screen_height_aspect);

		source_width = screen_width;
		source_height = screen_height;
		source_width_aspect = screen_width_aspect;
		source_height_aspect = screen_height_aspect;
#ifdef USE_SCREEN_ROTATE
	}
#ifdef _BML3MK5
	if (source_width != prev_source_width && source_height != prev_source_height) {
		HDC hdc = GetDC(main_window_handle);
		// for message board
		release_dib_section(hdcDibMixed, hBmpMixed, hOldBmpMixed, lpBufMixed);
		create_dib_section(hdc, source_width, source_height, &hdcDibMixed, &hBmpMixed, &hOldBmpMixed, &lpBufMixed, &lpBmpMixed, &lpDibMixed);
		ReleaseDC(main_window_handle, hdc);
	}
#endif
#endif

	if(config.stretch_screen && !window_mode && display_width >= source_width && display_height >= source_height) {
		if (!config.cutout_screen) {
			// fit to full screen
			stretched_width = (display_height * source_width_aspect) / source_height_aspect;
			stretched_height = display_height;
			if(stretched_width > display_width) {
				stretched_width = display_width;
				stretched_height = (display_width * source_height_aspect) / source_width_aspect;
			}
			display_stretched_width = display_width;
			display_stretched_height = display_height;

			screen_dest_x = (display_width - stretched_width) / 2;
			screen_dest_y = (display_height - stretched_height) / 2;
			screen_stretched_dest_x = screen_dest_x;
			screen_stretched_dest_y = screen_dest_y;

		} else {
			// fit text area to full screen
			stretched_width = (display_height * min_window_width) / min_window_height * screen_width / min_window_width;
			stretched_height = display_height * screen_height / min_window_height;
			if(stretched_width > display_width * screen_width / min_window_width) {
				stretched_width = display_width * screen_width / min_window_width;
				stretched_height = (display_width * min_window_height) / min_window_width * screen_height / min_window_height;
			}
			display_stretched_width = min_window_width;
			display_stretched_height = min_window_height;

			screen_dest_x = (display_width - stretched_width) / 2;
			screen_dest_y = (display_height - stretched_height) / 2;
			screen_stretched_dest_x = (min_window_width - screen_width) / 2;
			screen_stretched_dest_y = (min_window_height - screen_height) / 2;

		}
	}
	else {
//		int tmp_pow_x = display_width / source_width_aspect;
//		int tmp_pow_y = display_height / source_height_aspect;
//		int tmp_pow = 1;
//		if(tmp_pow_y >= tmp_pow_x && tmp_pow_x > 1) {
//			tmp_pow = tmp_pow_x;
//		}
//		else if(tmp_pow_x >= tmp_pow_y && tmp_pow_y > 1) {
//			tmp_pow = tmp_pow_y;
//		}
//		stretched_width = source_width_aspect * tmp_pow;
//		stretched_height = source_height_aspect * tmp_pow;
		stretched_width = source_width_aspect * power / 10;
		stretched_height = source_height_aspect * power / 10;
		display_stretched_width = display_width * 10 / power;
		display_stretched_height = display_height * 10 / power;

		screen_dest_x = (display_width - stretched_width) / 2;
		screen_dest_y = (display_height - stretched_height) / 2;
		screen_stretched_dest_x = screen_dest_x * 10 / power;
		screen_stretched_dest_y = screen_dest_y * 10 / power;
	}

	int new_pow_x = 1, new_pow_y = 1;
	while(stretched_width > source_width * new_pow_x) {
		new_pow_x++;
	}
	while(stretched_height > source_height * new_pow_y) {
		new_pow_y++;
	}

	// support high quality stretch only for x1 window size in gdi mode
	if(new_pow_x > 1 && new_pow_y > 1) {
		new_pow_x = new_pow_y = 1;
	}

//	if(stretched_width == source_width * new_pow_x && stretched_height == source_height * new_pow_y) {
//		new_pow_x = new_pow_y = 1;
//	}
	if(stretch_pow_x != new_pow_x || stretch_pow_y != new_pow_y) {
		stretch_pow_x = new_pow_x;
		stretch_pow_y = new_pow_y;
		stretch_changed = true;
	}

	if(stretch_changed) {
#ifdef USE_SMOOTH_STRETCH
		release_dib_section(hdcDibStretch1, hBmpStretch1, hOldBmpStretch1, lpBufStretch1);
		release_dib_section(hdcDibStretch2, hBmpStretch2, hOldBmpStretch2, lpBufStretch2);
#endif
		stretch_screen = false;
#ifdef USE_SMOOTH_STRETCH
		if(stretch_pow_x != 1 || stretch_pow_y != 1) {
			HDC hdc = GetDC(main_window_handle);
			create_dib_section(hdc, source_width * stretch_pow_x, source_height * stretch_pow_y, &hdcDibStretch1, &hBmpStretch1, &hOldBmpStretch1, &lpBufStretch1, &lpBmpStretch1, &lpDibStretch1);
			SetStretchBltMode(hdcDibStretch1, COLORONCOLOR);

			create_dib_section(hdc, stretched_width, stretched_height, &hdcDibStretch2, &hBmpStretch2, &hOldBmpStretch2, &lpBufStretch2, &lpBmpStretch2, &lpDibStretch2);
			SetStretchBltMode(hdcDibStretch2, HALFTONE);

			ReleaseDC(main_window_handle, hdc);
			stretch_screen = true;
		}
#endif
	}

//	first_draw_screen = false;
	first_draw_screen = true;
	first_invalidate = true;
	screen_size_changed = false;

	SetRect(&reD3Dmix, 0, 0, source_width, source_height);
	SetRect(&reD3Dsuf, screen_dest_x, screen_dest_y, screen_dest_x + stretched_width, screen_dest_y + stretched_height);
	if (screen_dest_x < 0) {
		reD3Dmix.left = -screen_stretched_dest_x; reD3Dmix.right = display_stretched_width - screen_stretched_dest_x;
		reD3Dsuf.left = 0;							reD3Dsuf.right = display_width;
	}
	if (screen_dest_y < 0) {
		reD3Dmix.top = -screen_stretched_dest_y; reD3Dmix.bottom = display_stretched_height - screen_stretched_dest_y;
		reD3Dsuf.top = 0;							reD3Dsuf.bottom = display_height;
	}
#ifdef _DEBUG_LOG
	out_debug(_T("set_display_size: w:%d h:%d p:%d mode:%s"),width,height,power,window_mode ? _T("true") : _T("false"));
	out_debug(_T("             mix: l:%d t:%d r:%d b:%d"),reD3Dmix.left,reD3Dmix.top,reD3Dmix.right,reD3Dmix.bottom);
	out_debug(_T("             suf: l:%d t:%d r:%d b:%d"),reD3Dsuf.left,reD3Dsuf.top,reD3Dsuf.right,reD3Dsuf.bottom);
#endif

#ifdef USE_DIRECT3D
	if (pD3Device) {
		HRESULT hre;

		if (window_mode) {
			d3dpp.BackBufferWidth = 0;
			d3dpp.BackBufferHeight = 0;
			d3dpp.Windowed = TRUE;
		} else {
			d3dpp.BackBufferWidth = display_width;
			d3dpp.BackBufferHeight = display_height;
			d3dpp.Windowed = TRUE;
		}
		// save buffer to dib temporary
		if (config.use_direct3d) {
			copy_d3dsuf_dib(pD3Dorigin, lpBmp, true);
		}
		// re create buffer
		hre = reset_d3device(main_window_handle);
		if (hre == D3D_OK) {
			pD3Device->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0,0,0), 0.0, 0);
			// restore buffer from dib
			copy_d3dsuf_dib(pD3Dorigin, lpBmp, false);
		}
		if (pD3Device != NULL) {
			LPDIRECT3DSURFACE9 suf = NULL;
			D3DSURFACE_DESC desc;
			pD3Device->GetBackBuffer(0,0,D3DBACKBUFFER_TYPE_MONO,&suf);
			if (suf != NULL) {
				suf->GetDesc(&desc);
				out_debug(_T("d3dbackbuffer: w:%d h:%d"),desc.Width,desc.Height);
				suf->Release();

				if (desc.Width < (UINT)(reD3Dsuf.right - reD3Dsuf.left)) {
					// adjust width to fit the surface
					uint32 new_mw = (uint32)((uint64)desc.Width * (reD3Dmix.right - reD3Dmix.left) / (reD3Dsuf.right - reD3Dsuf.left));
					uint32 new_sw = (uint32)((uint64)new_mw * (reD3Dsuf.right - reD3Dsuf.left) / (reD3Dmix.right - reD3Dmix.left));
					reD3Dmix.right = reD3Dmix.left + new_mw;
					reD3Dsuf.right = reD3Dsuf.left + new_sw;
				}
				if (desc.Height < (UINT)(reD3Dsuf.bottom - reD3Dsuf.top)) {
					// adjust height to fit the surface
					uint32 new_mh = (uint32)((uint64)desc.Height * (reD3Dmix.bottom - reD3Dmix.top) / (reD3Dsuf.bottom - reD3Dsuf.top));
					uint32 new_sh = (uint32)((uint64)new_mh * (reD3Dsuf.bottom - reD3Dsuf.top) / (reD3Dmix.bottom - reD3Dmix.top));
					reD3Dmix.bottom = reD3Dmix.top + new_mh;
					reD3Dsuf.bottom = reD3Dsuf.top + new_sh;
				}

				out_debug(_T("          mix: l:%d t:%d r:%d b:%d"),reD3Dmix.left,reD3Dmix.top,reD3Dmix.right,reD3Dmix.bottom);
				out_debug(_T("          suf: l:%d t:%d r:%d b:%d"),reD3Dsuf.left,reD3Dsuf.top,reD3Dsuf.right,reD3Dsuf.bottom);
			}
		}
#ifdef USE_SCREEN_ROTATE
		if(config.monitor_type) {
			pD3Dsource = pD3Drotate;
		} else
#endif
		{
			pD3Dsource = pD3Dorigin;
		}
	}
#endif	/* USE_DIRECT3D */

#ifdef _BML3MK5
	if (ledbox) {
		ledbox->SetMode(window_mode ? 0 : 1);
		ledbox->MoveInPlace();
		ledbox->SetPos(reD3Dmix.left, reD3Dmix.top);
	}
	if (msgboard) {
		msgboard->SetSize(source_width, source_height);
		msgboard->SetMessagePos(4+reD3Dmix.left, -4-reD3Dmix.top, 2);
		msgboard->SetInfoPos(-4-reD3Dmix.left, 4+reD3Dmix.top, 1);
	}
#endif
}

void EMU::change_screen_size(int sw, int sh, int swa, int sha, int ww, int wh)
{
	// virtual machine changes the screen size
	if(screen_width != sw || screen_height != sh) {
		screen_width = sw;
		screen_height = sh;
		screen_width_aspect = (swa != -1) ? swa : sw;
		screen_height_aspect = (sha != -1) ? sha : sh;
		window_width = ww;
		window_height = wh;
		screen_size_changed = true;

		// re-create dib sections
		HDC hdc = GetDC(main_window_handle);
		release_dib_section(hdcDib, hBmp, hOldBmp, lpBuf);
		create_dib_section(hdc, screen_width, screen_height, &hdcDib, &hBmp, &hOldBmp, &lpBuf, &lpBmp, &lpDib);
#ifdef USE_SCREEN_ROTATE
		release_dib_section(hdcDibRotate, hBmpRotate, hOldBmpRotate, lpBufRotate);
		create_dib_section(hdc, screen_height, screen_width, &hdcDibRotate, &hBmpRotate, &hOldBmpRotate, &lpBufRotate, &lpBmpRotate, &lpDibRotate);
#endif
#ifdef _BML3MK5
		// for message board
		release_dib_section(hdcDibMixed, hBmpMixed, hOldBmpMixed, lpBufMixed);
		create_dib_section(hdc, screen_width, screen_height, &hdcDibMixed, &hBmpMixed, &hOldBmpMixed, &lpBufMixed, &lpBmpMixed, &lpDibMixed);
#endif
		ReleaseDC(main_window_handle, hdc);

		// stop recording
		if(now_rec_vid) {
			stop_rec_video();
			stop_rec_sound();
		}

		// change the window size
		PostMessage(main_window_handle, WM_RESIZE, 0L, 0L);
	}
}

void EMU::draw_screen()
{
	// don't draw screen before new screen size is applied to buffers
	if(screen_size_changed) {
		return;
	}

#ifdef USE_DIRECT3D
	HRESULT hre;
	D3DLOCKED_RECT pLockedRect;

	lpD3DBmp = NULL;
	if (pD3Dorigin) {
		hre = pD3Dorigin->LockRect(&pLockedRect, NULL, 0);
		if (hre == D3D_OK) {
			lpD3DBmp = (scrntype *)pLockedRect.pBits;
		}
	}
#endif
	// draw screen
	vm->draw_screen();

#ifdef USE_DIRECT3D
	if (pD3Dorigin)	pD3Dorigin->UnlockRect();
#endif
	// screen size was changed in vm->draw_screen()
	if(screen_size_changed) {
		return;
	}

#ifdef USE_SCREEN_ROTATE
	// rotate screen
	// right turn
	if(config.monitor_type) {
#ifdef USE_DIRECT3D
		if (config.use_direct3d) {
			D3DLOCKED_RECT pLockedRectOrigin;
			D3DLOCKED_RECT pLockedRectRotate;
			hre = pD3Dorigin->LockRect(&pLockedRectOrigin, NULL, 0);
			hre = pD3Drotate->LockRect(&pLockedRectRotate, NULL, 0);
			for(int y = 0; y < screen_height; y++) {
				scrntype* src = (scrntype *)pLockedRectOrigin.pBits + screen_width * y;
				scrntype* out = (scrntype *)pLockedRectRotate.pBits + (screen_height - y - 1);
				for(int x = 0; x < screen_width; x++) {
					*out = src[x];
					out += screen_height;
				}
			}
			pD3Drotate->UnlockRect();
			pD3Dorigin->UnlockRect();
		} else
#endif
		{
			for(int y = 0; y < screen_height; y++) {
				scrntype* src = lpBmp + screen_width * (screen_height - y - 1);
				scrntype* out = lpBmpRotate + screen_height * (screen_width - 1) + (screen_height - y - 1);
				for(int x = 0; x < screen_width; x++) {
					*out = src[x];
					out -= screen_height;
				}
			}
		}
	}
#endif

#ifdef USE_DIRECT3D
	if (config.use_direct3d) {
		if (ledbox) {
			ledbox->Draw(pD3Dsource);
		}
	} else
#endif
	{
		if (ledbox) {
			ledbox->Draw(hdcDibSource);
		}
	}
}

void EMU::mix_screen()
{
#ifdef _BML3MK5
#ifdef USE_DIRECT3D
	if (pD3Device != NULL && config.use_direct3d) {
		pD3Device->UpdateSurface(pD3Dsource, NULL, pD3Dmixsuf, NULL);
		if (msgboard) {
			msgboard->Draw(pD3Dmixsuf);
		}
	} else
#endif
	{
		BitBlt(hdcDibMixed, 0, 0, source_width, source_height, hdcDibSource, 0, 0, SRCCOPY);
		if (msgboard) {
			msgboard->Draw(hdcDibMixed);
		}
	}
#endif

	// stretch screen
#ifdef USE_SMOOTH_STRETCH
	//	if(stretch_screen) {
	if(stretch_screen && (!config.use_direct3d)) {
#if 0
		StretchBlt(hdcDibStretch1, 0, 0, source_width * stretch_pow_x, source_height * stretch_pow_y, hdcDibSource, 0, 0, source_width, source_height, SRCCOPY);
#else
		// about 50% faster than StretchBlt()
		scrntype* src = lpBmpSource + source_width * (source_height - 1);
		scrntype* out = lpBmpStretch1 + source_width * stretch_pow_x * (source_height * stretch_pow_y - 1);
		int data_len = source_width * stretch_pow_x;

		for(int y = 0; y < source_height; y++) {
			if(stretch_pow_x != 1) {
				scrntype* out_tmp = out;
				for(int x = 0; x < source_width; x++) {
					scrntype c = src[x];
					for(int px = 0; px < stretch_pow_x; px++) {
						out_tmp[px] = c;
					}
					out_tmp += stretch_pow_x;
				}
			}
			else {
				// faster than memcpy()
				for(int x = 0; x < source_width; x++) {
					out[x] = src[x];
				}
			}
			if(stretch_pow_y != 1) {
				scrntype* src_tmp = out;
				for(int py = 1; py < stretch_pow_y; py++) {
					out -= data_len;
					// about 10% faster than memcpy()
					for(int x = 0; x < data_len; x++) {
						out[x] = src_tmp[x];
					}
				}
			}
			src -= source_width;
			out -= data_len;
		}
#endif

		StretchBlt(hdcDibStretch2, 0, 0, stretched_width, stretched_height, hdcDibStretch1, 0, 0, source_width * stretch_pow_x, source_height * stretch_pow_y, SRCCOPY);

	}
#endif
	first_draw_screen = true;

	// invalidate window
	InvalidateRect(main_window_handle, NULL, first_invalidate);
	UpdateWindow(main_window_handle);
	self_invalidate = true;
	skip_frame = false;

	// record picture
	if(now_rec_vid) {
#ifdef USE_DIRECT3D
		if (pD3Device != NULL && config.use_direct3d) {
			copy_d3dsuf_dib(pD3Dorigin, lpBmp, true);
		}
#endif
		copy_dib_rec_video();

		if(AVIStreamWrite(pAVICompressed, rec_frames++, 1, (LPBYTE)lpBmpRecVideo, pbmInfoHeader->biSizeImage, AVIIF_KEYFRAME, NULL, NULL) != AVIERR_OK) {
			out_log(LOG_ERROR,_T("AVIStreamWrite: Failed."));
			stop_rec_video();
		}
	}
}

scrntype* EMU::screen_buffer(int y)
{
#ifdef USE_DIRECT3D
	if (lpD3DBmp && config.use_direct3d) {
		return lpD3DBmp + screen_width * y;
	} else
#endif
	return lpBmp + screen_width * (screen_height - y - 1);
}

int EMU::screen_buffer_offset()
{
#ifdef USE_DIRECT3D
	if (lpD3DBmp && config.use_direct3d) {
		return screen_width;
	} else
#endif
	return -screen_width;
}

#ifdef USE_DIRECT3D
void EMU::update_screen_d3d()
{
	HRESULT hre = S_OK;
	PDIRECT3DSURFACE9 pD3Dsuf = NULL;

	if (lost_d3device) {
		// foCX
		hre = pD3Device->TestCooperativeLevel();
		if (hre == D3DERR_DEVICELOST) {
			// foCX ݃ZbgłȂ
			lost_d3device = true;
		} else if (hre == D3DERR_DEVICENOTRESET) {
			// foCX ݃Zbg\
			if (reset_d3device(main_window_handle) == D3D_OK) {
				lost_d3device = false;
			}
		} else {
			lost_d3device = false;
		}
	} else {
		if(first_draw_screen && pD3Device != NULL) {
			// obNobt@𓾂
			pD3Device->GetBackBuffer(0,0,D3DBACKBUFFER_TYPE_MONO,&pD3Dsuf);
			if (pD3Dsuf) {
				// ItCT[tF[XobNobt@w
				hre = pD3Device->StretchRect(pD3Dmixsuf, &reD3Dmix, pD3Dsuf, &reD3Dsuf, D3DTEXF_NONE);

				pD3Dsuf->Release();
			}
			// `
			hre = pD3Device->Present(NULL,NULL,NULL,NULL);

			first_invalidate = self_invalidate = false;
		}
		if (hre == D3DERR_DEVICELOST) {
			// foCX
			lost_d3device = true;
		}
	}

#ifdef _BML3MK5
	if (ledbox) {
		ledbox->Update(update_led());
	}
#endif
}
#endif

void EMU::update_screen(HDC hdc)
#ifdef USE_DIRECT3D // A{
{
	if (pD3Device != NULL && config.use_direct3d) {
		update_screen_d3d();
	} else {
		update_screen_dc(hdc);
	}
}

void EMU::update_screen_dc(HDC hdc)
#endif	// }A
{
#ifdef USE_BITMAP
	if(first_invalidate || !self_invalidate) {
		HDC hmdc = CreateCompatibleDC(hdc);
		HBITMAP hBitmap = LoadBitmap(instance_handle, _T("IDI_BITMAP1"));
		BITMAP bmp;
		GetObject(hBitmap, sizeof(BITMAP), &bmp);
		int w = (int)bmp.bmWidth;
		int h = (int)bmp.bmHeight;
		HBITMAP hOldBitmap = (HBITMAP)SelectObject(hmdc, hBitmap);
		BitBlt(hdc, 0, 0, w, h, hmdc, 0, 0, SRCCOPY);
		SelectObject(hmdc, hOldBitmap);
		DeleteObject(hBitmap);
		DeleteDC(hmdc);
	}
#endif
	if(first_draw_screen) {
#ifdef USE_LED
		// 7-seg LEDs
		for(int i = 0; i < MAX_LEDS; i++) {
			int x = leds[i].x;
			int y = leds[i].y;
			int w = leds[i].width;
			int h = leds[i].height;
			BitBlt(hdc, x, y, w, h, hdcDib, x, y, SRCCOPY);
		}
#else
		// standard screen
#ifdef USE_SMOOTH_STRETCH
		if(stretch_screen) {
			BitBlt(hdc, screen_dest_x, screen_dest_y, stretched_width, stretched_height, hdcDibStretch2, 0, 0, SRCCOPY);
		}
		else {
#endif
#ifdef _BML3MK5
			if(stretched_width == source_width && stretched_height == source_height) {
				BitBlt(hdc, screen_dest_x, screen_dest_y, stretched_width, stretched_height, hdcDibMixed, 0, 0, SRCCOPY);
			}
			else {
				StretchBlt(hdc, screen_dest_x, screen_dest_y, stretched_width, stretched_height, hdcDibMixed, 0, 0, source_width, source_height, SRCCOPY);
			}
#else
			if(stretched_width == source_width && stretched_height == source_height) {
				BitBlt(hdc, screen_dest_x, screen_dest_y, stretched_width, stretched_height, hdcDibSource, 0, 0, SRCCOPY);
			}
			else {
				StretchBlt(hdc, screen_dest_x, screen_dest_y, stretched_width, stretched_height, hdcDibSource, 0, 0, source_width, source_height, SRCCOPY);
			}
#endif
#ifdef USE_SMOOTH_STRETCH
		}
#endif

#ifdef USE_ACCESS_LAMP
		// draw access lamps of drives
		int status = vm->access_lamp() & 7;
		static int prev_status = 0;
		bool render_in = (status != 0);
		bool render_out = (prev_status != status);
		prev_status = status;

		if(render_in || render_out) {
			COLORREF crColor = RGB((status & 1) ? 255 : 0, (status & 2) ? 255 : 0, (status & 4) ? 255 : 0);
			int right_bottom_x = screen_dest_x + stretched_width;
			int right_bottom_y = screen_dest_y + stretched_height;

			for(int y = display_height - 6; y < display_height; y++) {
				for(int x = display_width - 6; x < display_width; x++) {
					if((x < right_bottom_x && y < right_bottom_y) ? render_in : render_out) {
						SetPixelV(hdc, x, y, crColor);
					}
				}
			}
		}
#endif
#endif
		first_invalidate = self_invalidate = false;
	}
#ifdef _BML3MK5
	if (ledbox) {
		ledbox->Update(update_led());
	}
#endif
}

#ifdef USE_DIRECT3D
void EMU::copy_d3dsuf_dib(PDIRECT3DSURFACE9 suf, scrntype *buf, bool to_dib)
{
	HRESULT hre;
	D3DLOCKED_RECT pLockedRect;

	hre = suf->LockRect(&pLockedRect, NULL, 0);
	if (hre == D3D_OK) {
		scrntype *src;
		scrntype *out;
		if (to_dib) {
			src = (scrntype *)pLockedRect.pBits;
			out = buf;
		} else {
			src = buf;
			out = (scrntype *)pLockedRect.pBits;
		}
		out += screen_width * (screen_height - 1);
		int data_len = screen_width;

		for(int y = 0; y < screen_height; y++) {
			for(int i = 0; i < data_len; i++) {
				out[i] = src[i];
			}
			src += data_len;
			out -= data_len;
		}
		suf->UnlockRect();
	}
}
#endif

void EMU::capture_screen()
{
#ifdef USE_DIRECT3D
	if (pD3Dorigin != NULL && config.use_direct3d) {
		copy_d3dsuf_dib(pD3Dorigin, lpBmp, true);
	}
#endif
	copy_dib_rec_video();

	_TCHAR app_path[_MAX_PATH], file_path[_MAX_PATH];
	application_path(app_path);
	_stprintf(file_path, _T("%s%04d-%02d-%02d_%02d-%02d-%02d.bmp"), app_path, sTime.wYear, sTime.wMonth, sTime.wDay, sTime.wHour, sTime.wMinute, sTime.wSecond);

	BITMAPFILEHEADER bmFileHeader = { (WORD)(TEXT('B') | TEXT('M') << 8) };
	bmFileHeader.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
	bmFileHeader.bfSize = bmFileHeader.bfOffBits + pbmInfoHeader->biSizeImage;

	DWORD dwSize;
	HANDLE hFile = CreateFile(file_path, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	WriteFile(hFile, &bmFileHeader, sizeof(BITMAPFILEHEADER), &dwSize, NULL);
	WriteFile(hFile, lpDibRecVideo, sizeof(BITMAPINFOHEADER), &dwSize, NULL);
	WriteFile(hFile, lpBmpRecVideo, pbmInfoHeader->biSizeImage, &dwSize, NULL);
	CloseHandle(hFile);
}

void EMU::start_rec_video(int fps, bool show_dialog)
{
	_TCHAR app_path[_MAX_PATH], file_path[_MAX_PATH];
	HRESULT res;

	application_path(app_path);
	_stprintf(file_path, _T("%svideo.avi"), app_path);

	// initialize vfw
	AVIFileInit();
	if((res = AVIFileOpen(&pAVIFile, file_path, OF_WRITE | OF_CREATE, NULL)) != AVIERR_OK) {
		out_log(LOG_ERROR,_T("AVIFileOpen: Failed. 0x%x"),res);
		return;
	}

	// stream header
	AVISTREAMINFO strhdr;
	memset(&strhdr, 0, sizeof(strhdr));
	strhdr.fccType = streamtypeVIDEO;	// vids
	strhdr.fccHandler = 0;
	strhdr.dwScale = 1;
	strhdr.dwRate = fps;
	strhdr.dwSuggestedBufferSize = pbmInfoHeader->biSizeImage;
//	SetRect(&strhdr.rcFrame, 0, 0, screen_width, screen_height);
	SetRect(&strhdr.rcFrame, 0, 0, rec_video_size[config.screen_video_size].w, rec_video_size[config.screen_video_size].h);
	if((res = AVIFileCreateStream(pAVIFile, &pAVIStream, &strhdr)) != AVIERR_OK) {
		out_log(LOG_ERROR,_T("AVIFileCreateStream: Failed. 0x%x"),res);
		stop_rec_video();
		return;
	}

	// compression
	AVICOMPRESSOPTIONS FAR * pOpts[1];
	pOpts[0] = &opts;
	if(show_dialog && !(AVISaveOptions(main_window_handle, ICMF_CHOOSE_KEYFRAME | ICMF_CHOOSE_DATARATE, 1, &pAVIStream, (LPAVICOMPRESSOPTIONS FAR *)&pOpts))) {
		AVISaveOptionsFree(1, (LPAVICOMPRESSOPTIONS FAR *)&pOpts);
		stop_rec_video();
		return;
	}
	if((res = AVIMakeCompressedStream(&pAVICompressed, pAVIStream, &opts, NULL)) != AVIERR_OK) {
		out_log(LOG_ERROR,_T("AVIMakeCompressedStream: Failed. 0x%x"),res);
		stop_rec_video();
		return;
	}
	if((res = AVIStreamSetFormat(pAVICompressed, 0, &lpDibRecVideo->bmiHeader, lpDibRecVideo->bmiHeader.biSize + lpDibSource->bmiHeader.biClrUsed * sizeof(RGBQUAD))) != AVIERR_OK) {
		out_log(LOG_ERROR,_T("AVIStreamSetFormat: Failed. 0x%x"),res);
		stop_rec_video();
		return;
	}
	rec_frames = 0;
	rec_fps = fps;
	now_rec_vid = true;
}

void EMU::stop_rec_video()
{
	// release vfw
	if(pAVIStream) {
		AVIStreamClose(pAVIStream);
	}
	if(pAVICompressed) {
		AVIStreamClose(pAVICompressed);
	}
	if(pAVIFile) {
		AVIFileClose(pAVIFile);
		AVIFileExit();
	}
	pAVIStream = NULL;
	pAVICompressed = NULL;
	pAVIFile = NULL;

	// repair header
	if(now_rec_vid) {
		_TCHAR app_path[_MAX_PATH], file_path[_MAX_PATH];
		application_path(app_path);
		_stprintf(file_path, _T("%svideo.avi"), app_path);

		FILE* fp = _tfopen(file_path, _T("r+b"));
		if(fp != NULL) {
			// copy fccHandler
			uint8 buf[4];
			fseek(fp, 0xbc, SEEK_SET);
			if(ftell(fp) == 0xbc) {
				fread(buf, 4, 1, fp);
				fseek(fp, 0x70, SEEK_SET);
				fwrite(buf, 4, 1, fp);
			}
			fclose(fp);
		}
	}
	now_rec_vid = false;
}

void EMU::restart_rec_video()
{
	if(now_rec_vid) {
		// release vfw
		if(pAVIStream) {
			AVIStreamClose(pAVIStream);
		}
		if(pAVICompressed) {
			AVIStreamClose(pAVICompressed);
		}
		if(pAVIFile) {
			AVIFileClose(pAVIFile);
			AVIFileExit();
		}
		pAVIStream = NULL;
		pAVICompressed = NULL;
		pAVIFile = NULL;

		start_rec_video(rec_fps, false);
	}
}

void EMU::resize_rec_video(int new_size)
{
#ifdef _BML3MK5
	if (config.screen_video_size != new_size) {
		release_dib_section(hdcDibRecVideo, hBmpRecVideo, hOldBmpRecVideo, lpBufRecVideo);
		HDC hdc = GetDC(main_window_handle);
		create_dib_rec_video(hdc, new_size);
		ReleaseDC(main_window_handle, hdc);
	}
	config.screen_video_size = new_size;

#endif
}

void EMU::create_dib_rec_video(HDC hdc, int new_size)
{
	if (new_size < 0 || new_size > 1) new_size = 0;
	create_dib_section(hdc, rec_video_size[new_size].w, rec_video_size[new_size].h, &hdcDibRecVideo, &hBmpRecVideo, &hOldBmpRecVideo, &lpBufRecVideo, &lpBmpRecVideo, &lpDibRecVideo);
	pbmInfoHeader = &lpDibRecVideo->bmiHeader;
}

void EMU::copy_dib_rec_video()
{
	BitBlt(hdcDibRecVideo, 0, 0,
		rec_video_size[config.screen_video_size].w, rec_video_size[config.screen_video_size].h,
		hdcDib, rec_video_size[config.screen_video_size].x, rec_video_size[config.screen_video_size].y, SRCCOPY);
}

//----------

int EMU::qsort_by_w_h_n(const void *a, const void *b)
{
	struct window_mode_st *a1 = (struct window_mode_st *)a;
	struct window_mode_st *b1 = (struct window_mode_st *)b;

	int rc = 0;
	if (!rc) rc = (a1->width - b1->width);
	if (!rc) rc = (a1->height - b1->height);
	if (!rc) rc = (a1->power - b1->power);
	return rc;
}

int EMU::qsort_by_n_w_h(const void *a, const void *b)
{
	struct screen_mode_st *a1 = (struct screen_mode_st *)a;
	struct screen_mode_st *b1 = (struct screen_mode_st *)b;

	int rc = 0;
	if (!rc) rc = (a1->disp_no - b1->disp_no);
	if (!rc) rc = (a1->width - b1->width);
	if (!rc) rc = (a1->height - b1->height);
	return rc;
}

BOOL CALLBACK EMU::MonitorProc(HMONITOR hMonitor, HDC hdcMonitor, LPRECT lprcMonitor, LPARAM dwData)
{
	EMU *emu = (EMU *)dwData;
	return emu->get_monitor_device(hMonitor, hdcMonitor, lprcMonitor);
}

BOOL EMU::get_monitor_device(HMONITOR hMonitor, HDC hdcMonitor, LPRECT lprcMonitor)
{
	if (monitor_count < disp_device_count) {
			disp_device[monitor_count].hMoni = hMonitor;
			CopyRect(&disp_device[monitor_count].re, lprcMonitor);
	}
	monitor_count++;

	return TRUE;
}

void EMU::enum_window_mode(int max_width, int max_height)
{
	// enumerate screen mode for window
	int w,h;
	bool minsize = false;
	bool maxsize = false;

	window_mode_count = 0;
	for (int i=0; window_proposal[i].cx != 0 && window_mode_count < WINDOW_MODE_MAX; i++) {
		w = window_proposal[i].cx;
		h = window_proposal[i].cy;
		// minimum window size?
		if (w < MIN_WINDOW_WIDTH || h < MIN_WINDOW_HEIGHT) {
			continue;
		}
		if (!minsize) {
			w = MIN_WINDOW_WIDTH;
			h = MIN_WINDOW_HEIGHT;
			i--;
			minsize = true;
		}
		if (w > MAX_WINDOW_WIDTH || h > MAX_WINDOW_HEIGHT) {
			if (!maxsize) {
				w = MAX_WINDOW_WIDTH;
				h = MAX_WINDOW_HEIGHT;
				maxsize = true;
				i--;
			} else {
				break;
			}
		}
		for (int pow = 10; pow <= 30 && window_mode_count < WINDOW_MODE_MAX; pow+=5) {
			if ((w*pow/10) <= max_width && (h*pow/10) <= max_height) {
				bool found = false;
				for(int j = 0; j < window_mode_count; j++) {
					if(window_mode[j].power == pow && window_mode[j].width == (w*pow/10) && window_mode[j].height == (h*pow/10)) {
						found = true;
						break;
					}
				}
				if (!found) {
					window_mode[window_mode_count].power = pow;
					window_mode[window_mode_count].width = w * pow / 10;
					window_mode[window_mode_count].height = h * pow / 10;
					out_debug(_T("window_mode:%d %dx%d x%.1f"),window_mode_count, w * pow / 10, h * pow / 10, (double)pow / 10.0);
					window_mode_count++;
				}
			} else {
				break;
			}
		}
	}
	qsort(&window_mode, window_mode_count, sizeof(struct window_mode_st), qsort_by_w_h_n);
}

void EMU::enum_screen_mode()
{
	// enumerate display device
	int max_desktop_width = 0;
	int max_desktop_height = 0;
	disp_device_count = 0;
	for(int i = 0; disp_device_count < DISP_DEVICE_MAX; i++) {
		DISPLAY_DEVICE disp;
		ZeroMemory(&disp, sizeof(disp));
		disp.cb = sizeof(disp);
		if (!EnumDisplayDevices(NULL, i, &disp, 0)) {
			break;
		}
		if (disp.StateFlags & (DISPLAY_DEVICE_ACTIVE | DISPLAY_DEVICE_PRIMARY_DEVICE)) {
			_tcsncpy(disp_device[disp_device_count].name, disp.DeviceName, 32);
			disp_device[disp_device_count].hMoni = NULL;
			disp_device_count++;
		}
	}

	// enumerate screen mode for fullscreen
	screen_mode_count = 0;
	for(int disp_no = 0; disp_no < disp_device_count && screen_mode_count < SCREEN_MODE_MAX; disp_no++) {
		for(int i = 0; screen_mode_count < SCREEN_MODE_MAX; i++) {
			DEVMODE dev;
			ZeroMemory(&dev, sizeof(dev));
			dev.dmSize = sizeof(dev);
			if(EnumDisplaySettings(disp_device[disp_no].name, i, &dev) == 0) {
				break;
			}
			if(((dev.dmPelsWidth * 30 / dev.dmPelsHeight) != 40)
			&& ((dev.dmPelsWidth * 40 / dev.dmPelsHeight) != 50)
			&& ((dev.dmPelsWidth * 90 / dev.dmPelsHeight) != 160)
			&& ((dev.dmPelsWidth * 50 / dev.dmPelsHeight) != 80)) {
				out_debug(_T("screen_mode:- [%d] %dx%d ignored"), disp_no, dev.dmPelsWidth, dev.dmPelsHeight);
				continue;
			}
			if (dev.dmPelsWidth <= 1280 && (dev.dmPelsWidth != 640 && dev.dmPelsWidth != 800 && dev.dmPelsWidth != 1024 &&  dev.dmPelsWidth != 1280)) {
				out_debug(_T("screen_mode:- [%d] %dx%d ignored"), disp_no, dev.dmPelsWidth, dev.dmPelsHeight);
				continue;
			}
			if(dev.dmPelsWidth >= MIN_WINDOW_WIDTH && dev.dmPelsHeight >= MIN_WINDOW_HEIGHT && dev.dmPelsWidth >= 640 && dev.dmPelsHeight >= 480) {
				bool found = false;
				for(int j = 0; j < screen_mode_count; j++) {
					if(screen_mode[j].disp_no == disp_no && screen_mode[j].width == dev.dmPelsWidth && screen_mode[j].height == dev.dmPelsHeight) {
						found = true;
						break;
					}
				}
				if(!found) {
					screen_mode[screen_mode_count].disp_no = disp_no;
					screen_mode[screen_mode_count].width = dev.dmPelsWidth;
					screen_mode[screen_mode_count].height = dev.dmPelsHeight;
					out_debug(_T("screen_mode:%d [%d] %dx%d"), screen_mode_count, disp_no, dev.dmPelsWidth, dev.dmPelsHeight);
					screen_mode_count++;

					if(max_desktop_width <= (int)dev.dmPelsWidth) {
						max_desktop_width = dev.dmPelsWidth;
					}
					if(max_desktop_height <= (int)dev.dmPelsHeight) {
						max_desktop_height = dev.dmPelsHeight;
					}
				}
			}
		}
	}
	qsort(&screen_mode, screen_mode_count, sizeof(struct screen_mode_st), qsort_by_n_w_h);

	if(screen_mode_count == 0) {
		screen_mode[screen_mode_count].disp_no = 0;
		screen_mode[screen_mode_count].width = desktop_width;
		screen_mode[screen_mode_count].height = desktop_height;
		screen_mode_count = 1;

		max_desktop_width = desktop_width;
		max_desktop_height = desktop_height;
	}

	// enumerate monitor size
	EnumDisplayMonitors(NULL, NULL, MonitorProc, (LPARAM)this);
}

void EMU::find_screen_mode()
{
	if (config.window_mode < 8) {
		bool find = false;
		int i = config.window_mode;
		if (i < window_mode_count) {
			if (window_mode[i].width == config.screen_width
			&&  window_mode[i].height == config.screen_height) {
				find = true;
				config.window_mode = i;
			}
		}
		if (!find) {
			config.window_mode = 0;
		}
	} else {
		bool find = false;
		for(int i = 0; i < screen_mode_count; i++) {
			if (screen_mode[i].disp_no == config.disp_device_no
			&&  screen_mode[i].width == config.screen_width
			&&  screen_mode[i].height == config.screen_height) {
				find = true;
				config.window_mode = i + 8;
				break;
			}
		}
		if (!find) {
			config.window_mode = 0;
		}
	}
}

void EMU::change_screen_mode(HWND hWnd, int mode)
{
//	if (mode == config.window_mode) return;

	if (mode == -1) {
		if (now_fullscreen) {
			mode = prev_window_mode;
		}
		prev_window_mode = config.window_mode;
	}

	set_window(hWnd, mode, desktop_width, desktop_height);
}

void EMU::change_stretch_screen()
{
	config.stretch_screen = !config.stretch_screen;
	set_display_size(-1, -1, 10, !now_fullscreen);
}

void EMU::change_cutout_screen()
{
	config.cutout_screen = !config.cutout_screen;
	set_display_size(-1, -1, 10, !now_fullscreen);
}

void EMU::set_window(HWND hWnd, int mode, int cur_width, int cur_height)
{
	static LONG style = WS_VISIBLE;
	WINDOWPLACEMENT place;
	place.length = sizeof(WINDOWPLACEMENT);

	if(mode >= 0 && mode < 8) {
		// go window
		if (mode >= window_mode_count) mode = 0;
//		int width = emu->get_window_width(mode);
//		int height = emu->get_window_height(mode);
		int width = window_mode[mode].width;
		int height = window_mode[mode].height;
#ifdef USE_SCREEN_ROTATE
		if (config.monitor_type) {
			int v = width;
			width = height;
			height = v;
		}
#endif
		RECT rect = {0, 0, width, height};
		AdjustWindowRectEx(&rect, dwStyle, TRUE, 0);
		int dest_x = (int)((desktop_width - (rect.right - rect.left)) / 2);
		int dest_y = (int)((desktop_height - (rect.bottom - rect.top)) / 2);
		//dest_x = (dest_x < 0) ? 0 : dest_x;
		dest_y = (dest_y < 0) ? 0 : dest_y;

		if(now_fullscreen) {
			ChangeDisplaySettings(NULL, 0);
			SetWindowLong(hWnd, GWL_STYLE, style);
			SetWindowPos(hWnd, HWND_TOP, dest_x, dest_y, rect.right - rect.left, rect.bottom - rect.top, SWP_SHOWWINDOW);
			now_fullscreen = false;
			// show menu
			show_menu_bar(hWnd);
		}
		else {
			SetWindowPos(hWnd, NULL, dest_x, dest_y, rect.right - rect.left, rect.bottom - rect.top, SWP_NOZORDER);
		}
		config.window_mode = prev_window_mode = mode;
		config.disp_device_no = 0;
		config.screen_width = width;
		config.screen_height = height;
		window_mode_power = window_mode[mode].power;

		// set screen size to emu class
		set_display_size(width, height, window_mode_power, true);
	}
	else if(!now_fullscreen) {
		// go fullscreen
		if (mode >= screen_mode_count + 8) mode = -1;
		int disp_no = (mode == -1) ? 0 : screen_mode[mode - 8].disp_no;
		int width = (mode == -1) ? cur_width : screen_mode[mode - 8].width;
		int height = (mode == -1) ? cur_height : screen_mode[mode - 8].height;
		int top = 0;
		int left = 0;

		DEVMODE dev;
		ZeroMemory(&dev, sizeof(dev));
		dev.dmSize = sizeof(dev);
		dev.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;
		dev.dmBitsPerPel = desktop_bpp;
		dev.dmPelsWidth = width;
		dev.dmPelsHeight = height;

		TCHAR *dev_name = NULL;
		if (disp_no < disp_device_count) {
			dev_name = disp_device[disp_no].name;
			top =  disp_device[disp_no].re.top;
			left = disp_device[disp_no].re.left;
		}

		if(ChangeDisplaySettingsEx(dev_name, &dev, NULL, CDS_TEST, NULL) == DISP_CHANGE_SUCCESSFUL) {
			GetWindowPlacement(hWnd, &place);
			ChangeDisplaySettingsEx(dev_name, &dev, NULL, CDS_FULLSCREEN, NULL);
			style = GetWindowLong(hWnd, GWL_STYLE);
			SetWindowLong(hWnd, GWL_STYLE, WS_VISIBLE);
			SetWindowPos(hWnd, HWND_TOP, left, top, width, height, SWP_SHOWWINDOW);
			SetCursorPos(width / 2, height / 2);
			now_fullscreen = true;

			if(mode == -1) {
				for(int i = 0; i < screen_mode_count; i++) {
					if(screen_mode[i].disp_no == disp_no && screen_mode[i].width == width && screen_mode[i].height == height) {
						mode = i + 8;
						break;
					}
				}
			}
			config.window_mode = mode;
			config.disp_device_no = disp_no;
			config.screen_width = width;
			config.screen_height = height;

			// remove menu
			hide_menu_bar(hWnd);

			// set screen size to emu class
			set_display_size(width, height, 10, false);
		}
	}

//	mute_sound(false);
	set_pause(1, false);
}

void EMU::get_window_mode_str(int num, _TCHAR *str)
{
	if (num < window_mode_count) {
		_stprintf(str, _T("%dx%d x%.1f"),
			window_mode[num].width,
			window_mode[num].height,
			(double)window_mode[num].power / 10.0);
	} else {
		_stprintf(str, _T(""));
	}
}

void EMU::get_screen_mode_str(int num, _TCHAR *str)
{
	if (num < screen_mode_count) {
		_stprintf(str, _T("[%d] %dx%d"),
			screen_mode[num].disp_no,
			screen_mode[num].width,
			screen_mode[num].height);
	} else {
		_stprintf(str, _T(""));
	}
}

void EMU::get_screen_record_size_str(int num, _TCHAR *str)
{
	if (0 <= num && num <= 1) {
		_stprintf(str, _T("%dx%d"),
			rec_video_size[num].w,
			rec_video_size[num].h);
	} else {
		_stprintf(str, _T(""));
	}
}

#ifdef USE_DIRECT3D
void EMU::initialize_d3device(HWND hWnd)
{
	ZeroMemory(&d3dpp , sizeof(D3DPRESENT_PARAMETERS));

	// D3D
	pD3D = Direct3DCreate9(D3D_SDK_VERSION);
	if (pD3D) {
		pD3D->GetAdapterDisplayMode(D3DADAPTER_DEFAULT , &d3ddm);

		d3dpp.BackBufferWidth = 0;
		d3dpp.BackBufferHeight = 0;
		d3dpp.BackBufferFormat = d3ddm.Format;
		d3dpp.BackBufferCount = 1;
		d3dpp.MultiSampleType = D3DMULTISAMPLE_NONE;
		d3dpp.MultiSampleQuality = 0;
		d3dpp.EnableAutoDepthStencil = FALSE;
		d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
		d3dpp.Windowed = TRUE;

		set_d3dpresent_interval();
	}
}

HRESULT EMU::create_d3device(HWND hWnd)
{
	D3DCAPS9 d3dcaps;
	HRESULT hre;

	lost_d3device = false;

	hre = pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd, D3DCREATE_HARDWARE_VERTEXPROCESSING ,&d3dpp, &pD3Device);
	if (hre != D3D_OK) {
		hre = pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd, D3DCREATE_SOFTWARE_VERTEXPROCESSING ,&d3dpp, &pD3Device);
	}
	if (hre != D3D_OK) {
		// disable direct 3d
		pD3Device = NULL;
	}
	enable_direct3d = false;

	if (pD3Device != NULL) {
		pD3Device->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0,0,0), 0.0, 0);

		pD3Device->GetDeviceCaps(&d3dcaps);
		if (d3dcaps.Caps == D3DCAPS_READ_SCANLINE && (d3dcaps.PresentationIntervals & D3DPRESENT_INTERVAL_ONE) != 0) {
			enable_direct3d = true;
		} else {
			config.use_direct3d = 0;
		}
	}
	return hre;
}

void EMU::create_d3dofflinesurface()
{
	HRESULT hre;
	if (pD3Device) {
		hre = pD3Device->CreateOffscreenPlainSurface(screen_width, screen_height, D3DFMT_X8R8G8B8, D3DPOOL_SYSTEMMEM, &pD3Dorigin, NULL);
		if (hre != D3D_OK) {
			release_d3device();
		}
		hre = pD3Device->CreateOffscreenPlainSurface(source_width, source_height, D3DFMT_X8R8G8B8, D3DPOOL_DEFAULT, &pD3Dmixsuf, NULL);
		if (hre != D3D_OK) {
			release_d3device();
		}
#ifdef USE_SCREEN_ROTATE
		hre = pD3Device->CreateOffscreenPlainSurface(screen_height, screen_width, D3DFMT_X8R8G8B8, D3DPOOL_SYSTEMMEM, &pD3Drotate, NULL);
		if (hre != D3D_OK) {
			release_d3device();
		}

		if(config.monitor_type) {
			pD3Dsource = pD3Drotate;
		} else
#endif
		{
			pD3Dsource = pD3Dorigin;
		}
	}
}

HRESULT EMU::reset_d3device(HWND hWnd)
{
	HRESULT hre = D3D_OK;

	hre = pD3Device->Reset(&d3dpp);
	if (hre == D3DERR_DEVICELOST) {
		return hre;
	}
	if (hre != D3D_OK) {
		// re create
		release_d3device();

		hre = create_d3device(hWnd);
		if (hre == D3D_OK) {
			create_d3dofflinesurface();
		}
	}
	return hre;
}

void EMU::release_d3device()
{
#ifdef USE_SCREEN_ROTATE
	if (pD3Drotate != NULL) {
		pD3Drotate->Release();
		pD3Drotate = NULL;
	}
#endif
	pD3Dsource = NULL;

	if (pD3Dorigin != NULL) {
		pD3Dorigin->Release();
		pD3Dorigin = NULL;
		lpD3DBmp = NULL;
	}
	if (pD3Dmixsuf != NULL) {
		pD3Dmixsuf->Release();
		pD3Dmixsuf = NULL;
	}
	if (pD3Device != NULL) {
		pD3Device->Release();
		pD3Device = NULL;
	}
}

void EMU::terminate_d3device()
{
	if (pD3D != NULL) {
		pD3D->Release();
	}
}

void EMU::set_d3dpresent_interval()
{
	if (config.use_direct3d == 2) {
		d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
	} else {
		d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_ONE;
	}
}

#endif
