/** @file win32_sound.cpp

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.08.18 -

	@brief [ win32 sound ]
*/

#include "emu.h"
#include "vm/vm.h"
#include "fileio.h"

#define DSOUND_BUFFER_SIZE (DWORD)(sound_samples * 8)
#define DSOUND_BUFFER_HALF (DWORD)(sound_samples * 4)

#define SOUND_RECORD_IN_MIXER

void EMU::initialize_sound(int rate, int samples, int latency)
{
	sound_rate = rate;
	sound_samples = samples;
	sound_latency_half = latency / 2;
	vm->initialize_sound(sound_rate, sound_samples);
	sound_ok = sound_started = now_mute = now_rec_snd = false;
	sound_prev_time = 0;

	// initialize direct sound
	PCMWAVEFORMAT pcmwf;
	DSBUFFERDESC dsbd;
	WAVEFORMATEX wfex;

	if(FAILED(DirectSoundCreate(NULL, &lpds, NULL))) {
		return;
	}
	if(FAILED(lpds->SetCooperativeLevel(main_window_handle, DSSCL_PRIORITY))) {
		return;
	}

	// primary buffer
	ZeroMemory(&dsbd, sizeof(dsbd));
	dsbd.dwSize = sizeof(dsbd);
	dsbd.dwFlags = DSBCAPS_PRIMARYBUFFER;
	if(FAILED(lpds->CreateSoundBuffer(&dsbd, &lpdsp, NULL))) {
		return;
	}
	ZeroMemory(&wfex, sizeof(wfex));
	wfex.wFormatTag = WAVE_FORMAT_PCM;
	wfex.nChannels = 2;
	wfex.wBitsPerSample = 16;
	wfex.nSamplesPerSec = sound_rate;
	wfex.nBlockAlign = wfex.nChannels * wfex.wBitsPerSample / 8;
	wfex.nAvgBytesPerSec = wfex.nSamplesPerSec * wfex.nBlockAlign;
	if(FAILED(lpdsp->SetFormat(&wfex))) {
		return;
	}

	// secondary buffer
	ZeroMemory(&pcmwf, sizeof(pcmwf));
	pcmwf.wf.wFormatTag = WAVE_FORMAT_PCM;
	pcmwf.wf.nChannels = 2;
	pcmwf.wBitsPerSample = 16;
	pcmwf.wf.nSamplesPerSec = sound_rate;
	pcmwf.wf.nBlockAlign = pcmwf.wf.nChannels * pcmwf.wBitsPerSample / 8;
	pcmwf.wf.nAvgBytesPerSec = pcmwf.wf.nSamplesPerSec * pcmwf.wf.nBlockAlign;
	ZeroMemory(&dsbd, sizeof(dsbd));
	dsbd.dwSize = sizeof(dsbd);
	dsbd.dwFlags = DSBCAPS_STICKYFOCUS | DSBCAPS_GETCURRENTPOSITION2;
	dsbd.dwBufferBytes = DSOUND_BUFFER_SIZE;
	dsbd.lpwfxFormat = (LPWAVEFORMATEX)&pcmwf;
	if(FAILED(lpds->CreateSoundBuffer(&dsbd, &lpdsb, NULL))) {
		return;
	}

	sound_ok = first_half = true;
}
void EMU::initialize_sound(int rate, int samples)
{
	initialize_sound(rate, samples, 100);
}

void EMU::initialize_sound()
{
	// load sound config
#ifdef SUPPORT_SOUND_FREQ_55467HZ
	// PC-8801/9801 series
	static int freq_table[8] = {2000, 4000, 8000, 11025, 22050, 44100, 55467, 96000};
#else
	static int freq_table[8] = {2000, 4000, 8000, 11025, 22050, 44100, 48000, 96000};
#endif
	static double late_table[5] = {0.05, 0.1, 0.2, 0.3, 0.4};

	if(!(0 <= config.sound_frequency && config.sound_frequency < 8)) {
		config.sound_frequency = 6;	// default: 48KHz
	}
	if(!(0 <= config.sound_latency && config.sound_latency < 5)) {
		config.sound_latency = 1;	// default: 100msec
	}
	int frequency = freq_table[config.sound_frequency];
	int samples = (int)((double)frequency * late_table[config.sound_latency] + 0.5);
	int latency = (int)(late_table[config.sound_latency] * 1000.0);

	initialize_sound(frequency, samples, latency);
}

void EMU::release_sound()
{
	// release direct sound
	if(lpdsp) {
		lpdsp->Release();
	}
	if(lpdsb) {
		lpdsb->Release();
	}
	if(lpds) {
		lpds->Release();
	}
	lpdsp = NULL;
	lpdsb = NULL;
	lpds = NULL;

	// stop recording
	stop_rec_sound();
}

void EMU::update_sound(int* extra_frames)
{
	*extra_frames = 0;

//	out_debug("EMU::update_sound %ld\n",timeGetTime());

	if(sound_ok) {
		DWORD play_c, write_c, offset, size1, size2;
		WORD *ptr1, *ptr2;

		// start play
		if(!sound_started) {
			lpdsb->Play(0, 0, DSBPLAY_LOOPING);
			sound_started = true;
			return;
		}

		// check current position
		if(FAILED(lpdsb->GetCurrentPosition(&play_c, &write_c))) {
			return;
		}
		if(first_half) {
			if(play_c < DSOUND_BUFFER_HALF) {
				return;
			}
			offset = 0;
		}
		else {
			if(play_c >= DSOUND_BUFFER_HALF) {
				return;
			}
			offset = DSOUND_BUFFER_HALF;
		}

		// skip create_sound at least once
		DWORD tim = timeGetTime();
		if ((int)(tim - sound_prev_time) < sound_latency_half) {
			return;
		}
		sound_prev_time = tim;

		// sound buffer must be updated
		uint16* sound_buffer = vm->create_sound(extra_frames);
#ifndef SOUND_RECORD_IN_MIXER
		record_sound(sound_buffer, sound_samples);
#endif
		if(lpdsb->Lock(offset, DSOUND_BUFFER_HALF, (void **)&ptr1, &size1, (void**)&ptr2, &size2, 0) == DSERR_BUFFERLOST) {
			lpdsb->Restore();
		}
//		out_debug("EMU::update_sound 1 tim:%ld playc:%ld size1:%ld size2:%ld",tim,play_c,size1,size2);
		if(sound_buffer) {
			if(ptr1) {
				CopyMemory(ptr1, sound_buffer, size1);
			}
			if(ptr2) {
				CopyMemory(ptr2, sound_buffer + size1, size2);
			}
		}
		lpdsb->Unlock(ptr1, size1, ptr2, size2);
		first_half = !first_half;
	}
}

void EMU::mute_sound(bool mute)
{
	if (mute) {
		if(!now_mute && sound_started) {
#if 1
			// sound is continuous driving during mute 
			// check current position
			DWORD size1, size2;
			WORD *ptr1, *ptr2;

			if(lpdsb->Lock(0, DSOUND_BUFFER_SIZE, (void **)&ptr1, &size1, (void**)&ptr2, &size2, 0) == DSERR_BUFFERLOST) {
				lpdsb->Restore();
			}
			if(ptr1) {
				ZeroMemory(ptr1, size1);
			}
			if(ptr2) {
				ZeroMemory(ptr2, size2);
			}
			lpdsb->Unlock(ptr1, size1, ptr2, size2);
#else
			lpdsb->Stop();
#endif
		}
		now_mute = true;
	} else {
		if(!vm_pause) {
			now_mute = false;
#if 0
			if (sound_started) {
				lpdsb->Play(0, 0, DSBPLAY_LOOPING);
			}
#endif
		}
	}
}

void EMU::set_volume(int volume)
{
//	if (lpdsb) {
//		lpdsb->SetVolume((LONG)volume);
//	}
	// set volume for sound devices.
	if (vm) {
		vm->set_volume();
	}
}


void EMU::start_rec_sound()
{
	if(!now_rec_snd) {
		_TCHAR app_path[_MAX_PATH], file_path[_MAX_PATH];
		application_path(app_path);
		_stprintf(file_path, _T("%ssound.wav"), app_path);

		rec = new FILEIO();
		if(rec->Fopen(file_path, FILEIO_WRITE_BINARY)) {
			// write dummy wave header
			wavheader_t header;

			memset(&header, 0, sizeof(wavheader_t));
			rec->Fwrite(&header, sizeof(wavheader_t), 1);
			rec_bytes = 0;
			now_rec_snd = true;
		}
		else {
			// failed to open the wave file
			delete rec;
		}
	}
}

void EMU::stop_rec_sound()
{
	if(now_rec_snd) {
		// update wave header
		wavheader_t header;

		header.dwRIFF = 0x46464952;
		header.dwFileSize = rec_bytes + sizeof(wavheader_t) - 8;
		header.dwWAVE = 0x45564157;
		header.dwfmt_ = 0x20746d66;
		header.dwFormatSize = 16;
		header.wFormatTag = 1;
		header.wChannels = 2;
		header.wBitsPerSample = 16;
		header.dwSamplesPerSec = sound_rate;
		header.wBlockAlign = header.wChannels * header.wBitsPerSample / 8;
		header.dwAvgBytesPerSec = header.dwSamplesPerSec * header.wBlockAlign;
		header.dwdata = 0x61746164;
		header.dwDataLength = rec_bytes;

		rec->Fseek(0, FILEIO_SEEK_SET);
		rec->Fwrite(&header, sizeof(wavheader_t), 1);
		rec->Fclose();

		delete rec;
		now_rec_snd = false;
	}
}

void EMU::restart_rec_sound()
{
	if(now_rec_snd) {
		rec->Fclose();
		delete rec;
		now_rec_snd = false;

		start_rec_sound();
	}
}

void EMU::record_rec_sound(uint16 *buffer, int samples)
{
	if(now_rec_snd) {
		// record sound
		int length = sizeof(uint16) * samples * 2;
		rec->Fwrite(buffer, length, 1);
		rec_bytes += length;
	}
}

void EMU::record_rec_sound(int32 *buffer, int samples)
{
#ifdef SOUND_RECORD_IN_MIXER
	if(now_rec_snd) {
		// record sound
		int length = sizeof(uint16) * samples * 2;
		uint16 dat;
		for(int i=0; i<(samples << 1); i++) {
			dat = (uint16)buffer[i];
			rec->Fwrite(&dat, sizeof(uint16), 1);
		}
		rec_bytes += length;
	}
#endif
}

//
//
//
void EMU::create_volume_dialog()
{
	// create volume dialog
	volumebox->Create(main_window_handle);
}
